import csv
import datetime
import threading
import traceback
from decimal import Decimal

from sqlalchemy import null, or_, select
from sqlalchemy.sql.expression import false

from werkzeug.exceptions import BadRequest

from zish import loads

from chellow.models import (
    Bill,
    BillType,
    Channel,
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    Era,
    GBill,
    GContract,
    GEra,
    GExitZone,
    GReadType,
    GReadingFrequency,
    GSupply,
    GUnit,
    GeneratorType,
    GspGroup,
    HhDatum,
    Llfc,
    MarketRole,
    MeterType,
    Participant,
    Party,
    Pc,
    ReadType,
    RegisterRead,
    Session,
    Site,
    Snag,
    Source,
    Ssc,
    Supply,
    Tpr,
    User,
    UserRole,
    VoltageLevel,
)
from chellow.utils import (
    hh_format,
    parse_bool,
    parse_channel_type,
    parse_hh_start,
    parse_mpan_core,
    parse_pc_code,
)

process_id = 0
process_lock = threading.Lock()
processes = {}

NO_CHANGE = "{no change}"

CHANNEL_TYPES = ("ACTIVE", "REACTIVE_IMP", "REACTIVE_EXP")


def add_arg(args, name, values, index):
    if index >= len(values):
        raise BadRequest(f"Another field called {name} needs to be added on the end.")

    value = values[index].strip()
    args.append((name, value))
    return value


ALLOWED_ACTIONS = ("insert", "update", "delete")


def general_import_era(sess, action, vals, args):
    if action == "update":
        mpan_core_str = add_arg(args, "mpan_core", vals, 0)
        mpan_core = parse_mpan_core(mpan_core_str)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        date_str = add_arg(args, "date", vals, 1)
        dt = parse_hh_start(date_str)
        era = supply.find_era_at(sess, dt)
        if era is None:
            raise BadRequest("There isn't a era at this date.")

        start_date_str = add_arg(args, "Start date", vals, 2)
        if start_date_str == NO_CHANGE:
            start_date = era.start_date
        else:
            start_date = parse_hh_start(start_date_str)

        finish_date_str = add_arg(args, "Finish date", vals, 3)
        if finish_date_str == NO_CHANGE:
            finish_date = era.finish_date
        else:
            finish_date = parse_hh_start(finish_date_str)

        mop_contract = None
        mop_contract_name = add_arg(args, "MOP Contract", vals, 4)
        if mop_contract_name == NO_CHANGE:
            mop_contract = era.mop_contract
        elif len(mop_contract_name) > 0:
            mop_contract = Contract.get_mop_by_name(sess, mop_contract_name)

        dc_contract = None
        dc_contract_name = add_arg(args, "DC Contract", vals, 5)
        if dc_contract_name == NO_CHANGE:
            dc_contract = era.dc_contract
        elif len(dc_contract_name) > 0:
            dc_contract = Contract.get_dc_by_name(sess, dc_contract_name)

        msn = add_arg(args, "Meter Serial Number", vals, 6)
        if msn == NO_CHANGE:
            msn = era.msn

        pc_code_raw = add_arg(args, "Profile Class", vals, 7)
        if pc_code_raw == NO_CHANGE:
            pc = era.pc
        else:
            pc_code = parse_pc_code(pc_code_raw)
            pc = Pc.get_by_code(sess, pc_code)

        mtc_code_raw = add_arg(args, "Meter Timeswitch Class", vals, 8)
        if mtc_code_raw == NO_CHANGE:
            mtc_code = era.mtc_participant.mtc.code
        else:
            mtc_code = mtc_code_raw

        cop_code = add_arg(args, "CoP", vals, 9)
        cop = era.cop if cop_code == NO_CHANGE else Cop.get_by_code(sess, cop_code)

        comm_code = add_arg(args, "Comms Type", vals, 10)
        comm = era.comm if comm_code == NO_CHANGE else Comm.get_by_code(sess, comm_code)

        ssc_code_str = add_arg(args, "SSC", vals, 11)
        if ssc_code_str == NO_CHANGE:
            ssc_code = None if era.ssc is None else era.ssc.code
        elif len(ssc_code_str) == 0:
            ssc_code = None
        else:
            ssc_code = ssc_code_str

        es_code = add_arg(args, "Energisation Status", vals, 12)
        if es_code == NO_CHANGE:
            es = era.energisation_status
        else:
            es = EnergisationStatus.get_by_code(sess, es_code)

        dtc_meter_type_code = add_arg(args, "DTC Meter Type", vals, 13)
        if dtc_meter_type_code == NO_CHANGE:
            dtc_meter_type = era.dtc_meter_type
        else:
            dtc_meter_type = DtcMeterType.get_by_code(sess, dtc_meter_type_code)

        imp_mpan_core = add_arg(args, "Import MPAN Core", vals, 14)
        imp_llfc_code = None
        imp_sc = None
        imp_supplier_contract = None
        imp_supplier_account = None
        if imp_mpan_core == NO_CHANGE:
            imp_mpan_core = era.imp_mpan_core
        elif len(imp_mpan_core) == 0:
            imp_mpan_core = None

        if imp_mpan_core is not None:
            imp_llfc_code = add_arg(args, "Import LLFC", vals, 15)
            if imp_llfc_code == NO_CHANGE:
                imp_llfc_code = era.imp_llfc.code

            imp_sc_str = add_arg(args, "Import Agreed Supply Capacity", vals, 16)
            if imp_sc_str == NO_CHANGE:
                imp_sc = era.imp_sc
            else:
                try:
                    imp_sc = int(imp_sc_str)
                except ValueError as e:
                    raise BadRequest(
                        f"The import agreed supply capacity must be an integer. {e}"
                    )

            imp_supplier_contract_name = add_arg(
                args, "Import Supplier Contract", vals, 17
            )
            if imp_supplier_contract_name == NO_CHANGE:
                imp_supplier_contract = era.imp_supplier_contract
            else:
                imp_supplier_contract = Contract.get_supplier_by_name(
                    sess, imp_supplier_contract_name
                )

            imp_supplier_account = add_arg(args, "Import Supplier Account", vals, 18)
            if imp_supplier_account == NO_CHANGE:
                imp_supplier_account = era.imp_supplier_account

        exp_mpan_core = None
        exp_llfc_code = None
        exp_supplier_contract = None
        exp_supplier_account = None
        exp_sc = None
        if len(vals) > 19:
            exp_mpan_core = add_arg(args, "Export MPAN Core", vals, 19)
            if exp_mpan_core == NO_CHANGE:
                exp_mpan_core = era.exp_mpan_core
            elif len(exp_mpan_core) == 0:
                exp_mpan_core = None

            if exp_mpan_core is not None:
                exp_llfc_code = add_arg(args, "Export LLFC", vals, 20)
                if exp_llfc_code == NO_CHANGE:
                    exp_llfc_code = era.exp_llfc.code

                exp_sc_str = add_arg(args, "Export Agreed Supply Capacity", vals, 21)
                if exp_sc_str == NO_CHANGE:
                    exp_sc = era.exp_sc
                else:
                    try:
                        exp_sc = int(exp_sc_str)
                    except ValueError as e:
                        raise BadRequest(
                            f"The export supply capacity must be an integer. {e}"
                        )

                exp_supplier_contract_name = add_arg(
                    args, "Export Supplier Contract", vals, 22
                )
                if exp_supplier_contract_name == NO_CHANGE:
                    exp_supplier_contract = era.exp_supplier_contract
                else:
                    exp_supplier_contract = Contract.get_supplier_by_name(
                        sess, exp_supplier_contract_name
                    )

                exp_supplier_account = add_arg(
                    args, "Export Supplier Account", vals, 23
                )
                if exp_supplier_account == NO_CHANGE:
                    exp_supplier_account = era.exp_supplier_account

        supply.update_era(
            sess,
            era,
            start_date,
            finish_date,
            mop_contract,
            dc_contract,
            msn,
            pc,
            mtc_code,
            cop,
            comm,
            ssc_code,
            es,
            dtc_meter_type,
            imp_mpan_core,
            imp_llfc_code,
            imp_supplier_contract,
            imp_supplier_account,
            imp_sc,
            era.imp_ca,
            exp_mpan_core,
            exp_llfc_code,
            exp_supplier_contract,
            exp_supplier_account,
            exp_sc,
            era.exp_ca,
        )
    elif action == "delete":
        mpan_core = add_arg(args, "MPAN Core", vals, 0)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        date_str = add_arg(args, "Date", vals, 1)
        dt = parse_hh_start(date_str)
        era = supply.find_era_at(sess, dt)
        if era is None:
            raise BadRequest("There isn't a era at this date.")

        supply.delete_era(sess, era)
    elif action == "insert":
        channel_set = set()
        mpan_core = add_arg(args, "MPAN Core", vals, 0)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        start_date_str = add_arg(args, "Start date", vals, 1)
        if len(start_date_str) == 0:
            start_date = None
        else:
            start_date = parse_hh_start(start_date_str)
        existing_era = (
            sess.query(Era)
            .filter(
                Era.supply == supply,
                or_(Era.finish_date == null(), Era.finish_date >= start_date),
            )
            .order_by(Era.start_date)
            .first()
        )
        if existing_era is None:
            raise BadRequest("The start date is after end of the supply.")

        site_code = add_arg(args, "Site Code", vals, 2)
        physical_site = None
        logical_sites = []
        if site_code == NO_CHANGE:
            for site_era in existing_era.site_eras:
                if site_era.is_physical:
                    physical_site = site_era.site
                else:
                    logical_sites.append(site_era.site)
        else:
            physical_site = Site.get_by_code(sess, site_code)

        mop_contract_name = add_arg(args, "MOP Contract", vals, 3)
        if mop_contract_name == NO_CHANGE:
            mop_contract = existing_era.mop_contract
        else:
            mop_contract = Contract.get_mop_by_name(sess, mop_contract_name)

        dc_contract_name = add_arg(args, "DC Contract", vals, 4)
        if dc_contract_name == NO_CHANGE:
            dc_contract = existing_era.dc_contract
        else:
            dc_contract = Contract.get_dc_by_name(sess, dc_contract_name)

        msn = add_arg(args, "Meter Serial Number", vals, 5)
        if msn == NO_CHANGE:
            msn = existing_era.msn

        pc_code_raw = add_arg(args, "Profile Class", vals, 6)
        if pc_code_raw == NO_CHANGE:
            pc = existing_era.pc
        else:
            pc_code = parse_pc_code(pc_code_raw)
            pc = Pc.get_by_code(sess, pc_code)

        mtc_code_raw = add_arg(args, "Meter Timeswitch Class", vals, 7)
        if mtc_code_raw == NO_CHANGE:
            mtc_code = existing_era.mtc_participant.mtc.code
        else:
            mtc_code = mtc_code_raw

        cop_code = add_arg(args, "CoP", vals, 8)
        if cop_code == NO_CHANGE:
            cop = existing_era.cop
        else:
            cop = Cop.get_by_code(sess, cop_code)

        comm_code = add_arg(args, "Comms Type", vals, 9)
        if comm_code == NO_CHANGE:
            comm = existing_era.comm
        else:
            comm = Comm.get_by_code(sess, comm_code)

        ssc_code = add_arg(args, "Standard Settlement Configuration", vals, 10)
        if ssc_code == NO_CHANGE:
            ssc = existing_era.ssc
            if ssc is None:
                ssc_code = None
            else:
                ssc_code = ssc.code
        elif len(ssc_code) > 0:
            Ssc.get_by_code(sess, ssc_code, start_date)
        else:
            ssc_code = None

        es_code = add_arg(args, "Energisation Status", vals, 11)
        if es_code == NO_CHANGE:
            energisation_status = existing_era.energisation_status
        else:
            energisation_status = EnergisationStatus.get_by_code(sess, es_code)

        dtc_meter_type_code = add_arg(args, "DTC Meter Type", vals, 12)
        if dtc_meter_type_code == NO_CHANGE:
            dtc_meter_type = existing_era.dtc_meter_type
        else:
            dtc_meter_type = DtcMeterType.get_by_code(sess, dtc_meter_type_code)

        imp_mpan_core = add_arg(args, "Import MPAN Core", vals, 13)
        if imp_mpan_core == NO_CHANGE:
            imp_mpan_core = existing_era.imp_mpan_core
        elif len(imp_mpan_core) == 0:
            imp_mpan_core = None

        imp_supplier_contract = None
        imp_supplier_account = None
        imp_sc = None
        imp_llfc_code = None
        imp_supplier_contract_name = None

        if imp_mpan_core is not None:
            imp_llfc_code = add_arg(args, "Import Line Loss Factor Class", vals, 14)
            if imp_llfc_code == NO_CHANGE and existing_era.imp_llfc is not None:
                imp_llfc_code = existing_era.imp_llfc.code

            imp_sc_str = add_arg(args, "Import Agreed Supply Capacity", vals, 15)
            if imp_sc_str == NO_CHANGE:
                imp_sc = existing_era.imp_sc
            else:
                try:
                    imp_sc = int(imp_sc_str)
                except ValueError as e:
                    raise BadRequest(
                        f"The import agreed supply capacity must be an integer. {e}"
                    )

            imp_supplier_contract_name = add_arg(
                args, "Import Supplier Contract", vals, 16
            )
            if imp_supplier_contract_name == NO_CHANGE:
                imp_supplier_contract = existing_era.imp_supplier_contract
            else:
                imp_supplier_contract = Contract.get_supplier_by_name(
                    sess, imp_supplier_contract_name
                )

            imp_supplier_account = add_arg(
                args, "Import Supplier Account Reference", vals, 17
            )
            if imp_supplier_account == NO_CHANGE:
                imp_supplier_account = existing_era.imp_supplier_account

            for i, ctype in enumerate(CHANNEL_TYPES):
                field_name = f"Import {ctype}?"
                has_chan_str = add_arg(args, field_name, vals, i + 18)
                if has_chan_str == NO_CHANGE:
                    if existing_era.find_channel(sess, True, ctype) is not None:
                        channel_set.add((True, ctype))
                elif parse_bool(has_chan_str):
                    channel_set.add((True, ctype))

        exp_mpan_core = None
        exp_llfc_code = None
        exp_supplier_contract = None
        exp_supplier_account = None
        exp_sc = None

        if len(vals) > 21:
            exp_mpan_core = add_arg(args, "Export MPAN", vals, 21)
            if exp_mpan_core == NO_CHANGE:
                exp_mpan_core = existing_era.exp_mpan_core
            elif len(exp_mpan_core) == 0:
                exp_mpan_core = None

            if exp_mpan_core is not None:
                exp_llfc_code = add_arg(args, "Export LLFC", vals, 22)
                if exp_llfc_code == NO_CHANGE and existing_era.exp_llfc is not None:
                    exp_llfc_code = existing_era.exp_llfc.code

                exp_sc_str = add_arg(args, "Export Agreed Supply Capacity", vals, 23)
                if exp_sc_str == NO_CHANGE:
                    exp_sc = existing_era.exp_sc
                else:
                    try:
                        exp_sc = int(exp_sc_str)
                    except ValueError as e:
                        raise BadRequest(
                            f"The export supply capacity must be an integer. {e}"
                        )

                exp_supplier_contract_name = add_arg(
                    args, "Export Supplier Contract", vals, 24
                )
                if exp_supplier_contract_name == NO_CHANGE:
                    exp_supplier_contract = existing_era.exp_supplier_contract
                else:
                    exp_supplier_contract = Contract.get_supplier_by_name(
                        sess, exp_supplier_contract_name
                    )

                exp_supplier_account = add_arg(
                    args, "Export Supplier Account", vals, 25
                )
                if exp_supplier_account == NO_CHANGE:
                    exp_supplier_account = existing_era.exp_supplier_account

                for i, ctype in enumerate(CHANNEL_TYPES):
                    field_name = f"Export {ctype}?"
                    has_chan_str = add_arg(args, field_name, vals, i + 26)
                    if has_chan_str == NO_CHANGE:
                        if existing_era.find_channel(sess, False, ctype) is not None:
                            channel_set.add((False, ctype))
                    elif parse_bool(has_chan_str):
                        channel_set.add((False, ctype))

        supply.insert_era(
            sess,
            physical_site,
            logical_sites,
            start_date,
            None,
            mop_contract,
            dc_contract,
            msn,
            pc,
            mtc_code,
            cop,
            comm,
            ssc_code,
            energisation_status,
            dtc_meter_type,
            imp_mpan_core,
            imp_llfc_code,
            imp_supplier_contract,
            imp_supplier_account,
            imp_sc,
            existing_era.imp_ca,
            exp_mpan_core,
            exp_llfc_code,
            exp_supplier_contract,
            exp_supplier_account,
            exp_sc,
            existing_era.exp_ca,
            channel_set,
        )


def general_import_g_supply(sess, action, vals, args):
    if action == "insert":
        site_code = add_arg(args, "Site Code", vals, 0)
        site = Site.get_by_code(sess, site_code)
        mprn = add_arg(args, "MPRN", vals, 1)
        supply_name = add_arg(args, "Supply Name", vals, 2)
        g_exit_zone_code = add_arg(args, "Exit Zone", vals, 3)
        g_exit_zone = GExitZone.get_by_code(sess, g_exit_zone_code)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        msn = add_arg(args, "Meter Serial Number", vals, 6)
        correction_factor_str = add_arg(args, "Correction Factor", vals, 7)
        correction_factor = Decimal(correction_factor_str)
        g_unit_code = add_arg(args, "Unit of Measurement", vals, 8)
        g_unit = GUnit.get_by_code(sess, g_unit_code)
        g_contract_name = add_arg(args, "Supplier Contract", vals, 9)
        if len(g_contract_name) > 0:
            g_contract = GContract.get_supplier_by_name(sess, g_contract_name)
        else:
            g_contract = None
        account = add_arg(args, "Account", vals, 10)
        g_reading_frequency_code = add_arg(args, "Reading Frequency", vals, 11)
        g_reading_frequency = GReadingFrequency.get_by_code(
            sess, g_reading_frequency_code
        )
        aq_str = add_arg(args, "AQ", vals, 12)
        aq = Decimal(aq_str)
        soq_str = add_arg(args, "SOQ", vals, 13)
        soq = Decimal(soq_str)

        site.insert_g_supply(
            sess,
            mprn,
            supply_name,
            g_exit_zone,
            start_date,
            finish_date,
            msn,
            correction_factor,
            g_unit,
            g_contract,
            account,
            g_reading_frequency,
            aq,
            soq,
        )
        sess.flush()
    elif action == "update":
        existing_mprn = add_arg(args, "Existing MPRN", vals, 0)
        g_supply = GSupply.get_by_mprn(sess, existing_mprn)
        mprn = add_arg(args, "MPRN", vals, 0)
        mprn = g_supply.mprn if mprn == NO_CHANGE else mprn
        g_supply_name = add_arg(args, "Supply Name", vals, 4)
        if g_supply_name == NO_CHANGE:
            g_supply_name = g_supply.name
        g_supply.update(mprn, supply_name)
        sess.flush()
    elif action == "delete":
        mprn = add_arg(args, "MPRN", vals, 0)
        g_supply = GSupply.get_by_mprn(sess, mprn)
        g_supply.delete()
        sess.flush()


def general_import_g_era(sess, action, vals, args):
    if action == "insert":
        mprn = add_arg(args, "MPRN", vals, 0)
        g_supply = GSupply.get_by_mprn(sess, mprn)
        start_date_str = add_arg(args, "Start Date", vals, 1)
        if len(start_date_str) == 0:
            start_date = None
        else:
            start_date = parse_hh_start(start_date_str)

        existing_g_era = sess.scalars(
            select(GEra).where(
                GEra.g_supply == g_supply,
                GEra.start_date <= start_date,
                or_(GEra.finish_date == null(), GEra.finish_date >= start_date),
            )
        ).one_or_none()
        if existing_g_era is None:
            raise BadRequest("The start date is outside the dates of the supply.")

        site_code = add_arg(args, "Site Code", vals, 2)
        physical_site = None
        logical_sites = []
        if site_code == NO_CHANGE:
            for site_g_era in existing_g_era.site_g_eras:
                if site_g_era.is_physical:
                    physical_site = site_g_era.site
                else:
                    logical_sites.append(site_g_era.site)
        else:
            physical_site = Site.get_by_code(sess, site_code)

        msn = add_arg(args, "Meter Serial Number", vals, 3)
        if msn == NO_CHANGE:
            msn = existing_g_era.msn

        correction_factor_str = add_arg(args, "Correction Factor", vals, 4)
        if correction_factor_str == NO_CHANGE:
            correction_factor = existing_g_era.correction_factor
        else:
            correction_factor = Decimal(correction_factor_str)

        g_unit_code = add_arg(args, "Unit", vals, 5)
        if g_unit_code == NO_CHANGE:
            g_unit = existing_g_era.g_unit
        else:
            g_unit = GUnit.get_by_code(sess, g_unit_code)

        contract_name = add_arg(args, "Supplier Contract Name", vals, 6)
        if contract_name == NO_CHANGE:
            g_contract = existing_g_era.g_contract
        else:
            g_contract = GContract.get_supplier_by_name(sess, contract_name)

        account = add_arg(args, "Account", vals, 7)
        if account == NO_CHANGE:
            account = existing_g_era.account

        g_reading_frequency_code = add_arg(args, "Reading Frequency", vals, 8)
        if g_reading_frequency_code == NO_CHANGE:
            g_reading_frequency = existing_g_era.g_reading_frequency
        else:
            g_reading_frequency = GReadingFrequency.get_by_code(
                sess, g_reading_frequency_code
            )

        aq_str = add_arg(args, "AQ", vals, 9)
        aq = existing_g_era.aq if aq_str == NO_CHANGE else Decimal(aq_str)

        soq_str = add_arg(args, "SOQ", vals, 10)
        soq = existing_g_era.soq if soq_str == NO_CHANGE else Decimal(soq_str)

        g_supply.insert_g_era(
            sess,
            physical_site,
            logical_sites,
            start_date,
            None,
            msn,
            correction_factor,
            g_unit,
            g_contract,
            account,
            g_reading_frequency,
            aq,
            soq,
        )
        sess.flush()

    elif action == "update":
        mprn = add_arg(args, "MPRN", vals, 0)
        g_supply = GSupply.get_by_mprn(sess, mprn)
        date_str = add_arg(args, "date", vals, 1)
        dt = parse_hh_start(date_str)
        g_era = g_supply.find_g_era_at(sess, dt)
        if g_era is None:
            raise BadRequest("There isn't a gas era at this date.")

        start_date_str = add_arg(args, "Start Date", vals, 2)
        if start_date_str == NO_CHANGE:
            start_date = g_era.start_date
        else:
            start_date = parse_hh_start(start_date_str)

        finish_date_str = add_arg(args, "Finish Date", vals, 3)
        if finish_date_str == NO_CHANGE:
            finish_date = g_era.finish_date
        else:
            finish_date = parse_hh_start(finish_date_str)

        msn = add_arg(args, "Meter Serial Number", vals, 4)
        if msn == NO_CHANGE:
            msn = g_era.msn

        correction_factor_str = add_arg(args, "Correction Factor", vals, 5)
        if correction_factor_str == NO_CHANGE:
            correction_factor = g_era.correction_factor
        else:
            correction_factor = Decimal(correction_factor_str)

        g_unit_code = add_arg(args, "Unit", vals, 6)
        if g_unit_code == NO_CHANGE:
            g_unit = g_era.g_unit
        else:
            g_unit = GUnit.get_by_code(sess, g_unit_code)

        contract_name = add_arg(args, "Supplier Contract Name", vals, 7)
        if contract_name == NO_CHANGE:
            g_contract = g_era.g_contract
        else:
            g_contract = GContract.get_supplier_by_name(sess, contract_name)

        account = add_arg(args, "Account", vals, 8)
        if account == NO_CHANGE:
            account = g_era.account

        g_reading_frequency_code = add_arg(args, "Reading Frequency", vals, 9)
        if g_reading_frequency_code == NO_CHANGE:
            g_reading_frequency = g_era.g_reading_frequency
        else:
            g_reading_frequency = GReadingFrequency.get_by_code(
                sess, g_reading_frequency_code
            )

        aq_str = add_arg(args, "AQ", vals, 10)
        aq = g_era.aq if aq_str == NO_CHANGE else Decimal(aq_str)

        soq_str = add_arg(args, "SOQ", vals, 11)
        soq = g_era.aq if soq_str == NO_CHANGE else Decimal(soq_str)

        g_supply.update_g_era(
            sess,
            g_era,
            start_date,
            finish_date,
            msn,
            correction_factor,
            g_unit,
            g_contract,
            account,
            g_reading_frequency,
            aq,
            soq,
        )
    elif action == "delete":
        mprn = add_arg(args, "MPRN", vals, 0)
        g_supply = GSupply.get_by_mprn(sess, mprn)
        date_str = add_arg(args, "Date", vals, 1)
        dt = parse_hh_start(date_str)
        g_era = g_supply.find_g_era_at(sess, dt)
        if g_era is None:
            raise BadRequest("There isn't a gas era at this date.")

        g_supply.delete_g_era(sess, g_era)
        sess.flush()


def general_import_g_batch(sess, action, vals, args):
    if action == "insert":
        contract_name = add_arg(args, "Supplier Contract Name", vals, 0)
        g_contract = GContract.get_supplier_by_name(sess, contract_name)

        reference = add_arg(args, "Reference", vals, 1)
        description = add_arg(args, "Description", vals, 2)
        g_contract.insert_g_batch(sess, reference, description)

    elif action == "update":
        contract_name = add_arg(args, "Supplier Contract Name", vals, 0)
        g_contract = GContract.get_supplier_by_name(sess, contract_name)

        old_reference = add_arg(args, "Old Reference", vals, 1)
        g_batch = g_contract.get_batch(sess, old_reference)
        new_reference = add_arg(args, "New Reference", vals, 2)
        description = add_arg(args, "Description", vals, 3)
        g_batch.update(sess, new_reference, description)


def general_import_party(sess, action, vals, args):
    if action == "insert":
        market_role_code = add_arg(args, "Market Role Code", vals, 0)
        market_role = MarketRole.get_by_code(sess, market_role_code)
        participant_code = add_arg(args, "Participant Code", vals, 1)
        participant = Participant.get_by_code(sess, participant_code)
        name = add_arg(args, "Name", vals, 2)
        valid_from_str = add_arg(args, "Valid From", vals, 3)
        valid_from = parse_hh_start(valid_from_str)
        valid_to_str = add_arg(args, "Valid To", vals, 4)
        valid_to = parse_hh_start(valid_to_str)
        dno_code_str = add_arg(args, "DNO Code", vals, 5)
        dno_code = None if len(dno_code_str) == 0 else dno_code_str
        party = Party(
            market_role=market_role,
            participant=participant,
            name=name,
            valid_from=valid_from,
            valid_to=valid_to,
            dno_code=dno_code,
        )
        sess.add(party)
        sess.flush()

    elif action == "update":
        market_role_code = add_arg(args, "Market Role Code", vals, 0)
        participant_code = add_arg(args, "Participant Code", vals, 1)
        party = Party.get_by_participant_code_role_code(
            sess, participant_code, market_role_code
        )
        name = add_arg(args, "Name", vals, 2)
        party.name = name
        valid_from_str = add_arg(args, "Valid From", vals, 3)
        party.valid_from = parse_hh_start(valid_from_str)
        valid_to_str = add_arg(args, "Valid To", vals, 4)
        party.valid_to = parse_hh_start(valid_to_str)
        dno_code_str = add_arg(args, "DNO Code", vals, 5)
        dno_code = None if len(dno_code_str) == 0 else dno_code_str
        party.dno_code = dno_code
        sess.flush()


def general_import_meter_type(sess, action, vals, args):
    if action == "insert":
        code = add_arg(args, "Code", vals, 0)
        description = add_arg(args, "Description", vals, 1)
        valid_from_str = add_arg(args, "Valid From", vals, 2)
        valid_from = parse_hh_start(valid_from_str)
        valid_to_str = add_arg(args, "Valid To", vals, 3)
        valid_to = parse_hh_start(valid_to_str)
        meter_type = MeterType(
            code=code, description=description, valid_from=valid_from, valid_to=valid_to
        )
        sess.add(meter_type)
        sess.flush()

    elif action == "update":
        code = add_arg(args, "Code", vals, 0)
        mt = sess.query(MeterType).filter(MeterType.code == code).first()
        description = add_arg(args, "Description", vals, 1)
        mt.description = description
        valid_from_str = add_arg(args, "Valid From", vals, 2)
        valid_from = parse_hh_start(valid_from_str)
        mt.valid_from = valid_from
        valid_to_str = add_arg(args, "Valid To", vals, 3)
        valid_to = parse_hh_start(valid_to_str)
        mt.valid_to = valid_to
        sess.flush()


def _parse_breakdown(breakdown_str):
    if len(breakdown_str) == 0:
        return {}
    else:
        try:
            return loads(breakdown_str)
        except SyntaxError as e:
            raise BadRequest("Problem parsing the breakdown: " + str(e))


def _truncate_vals(vals):
    return _truncate_vals(vals[:-1]) if len(vals) > 0 and vals[0] == "" else vals


def general_import_mop_bill(sess, action, vals, args):
    vals = _truncate_vals(vals)
    if action == "insert":
        contract_name = add_arg(args, "Contract Name", vals, 0)
        contract = Contract.get_mop_by_name(sess, contract_name)
        batch_reference = add_arg(args, "Batch Reference", vals, 1)
        batch = contract.get_batch(sess, batch_reference)
        mpan_core_str = add_arg(args, "Mpan Core", vals, 2)
        mpan_core = parse_mpan_core(mpan_core_str)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        issue_date_str = add_arg(args, "Issue Date", vals, 3)
        issue_date = parse_hh_start(issue_date_str)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        net_str = add_arg(args, "Net", vals, 6)
        net = Decimal(net_str)
        vat_str = add_arg(args, "Vat", vals, 7)
        vat = Decimal(vat_str)
        gross_str = add_arg(args, "Gross", vals, 8)
        gross = Decimal(gross_str)
        reference = add_arg(args, "Reference", vals, 9)
        type_code = add_arg(args, "Type", vals, 10)
        typ = BillType.get_by_code(sess, type_code)
        breakdown_str = add_arg(args, "Breakdown", vals, 11)
        breakdown = _parse_breakdown(breakdown_str)

        bill = batch.insert_bill(
            sess,
            "",
            reference,
            issue_date,
            start_date,
            finish_date,
            0,
            net,
            vat,
            gross,
            typ,
            breakdown,
            supply,
        )

    elif action == "update":
        bill_id_str = add_arg(args, "Bill Id", vals, 0)
        bill_id = int(bill_id_str)
        bill = Bill.get_by_id(bill_id)

        reference = add_arg(args, "Reference", vals, 1)
        if reference == NO_CHANGE:
            reference = bill.getReference()

        issue_date_str = add_arg(args, "Issue Date", vals, 2)
        issue_date = None

        if issue_date_str == NO_CHANGE:
            issue_date = bill.issue_date
        else:
            issue_date = parse_hh_start(issue_date_str)

        start_date_str = add_arg(args, "Start Date", vals, 3)
        if start_date_str == NO_CHANGE:
            start_date = bill.start_date
        else:
            start_date = parse_hh_start(start_date_str)

        finish_date_str = add_arg(args, "Finish Date", vals, 4)
        if finish_date_str == NO_CHANGE:
            finish_date = bill.finish_date
        else:
            finish_date = parse_hh_start(finish_date_str)

        kwh_str = add_arg(args, "kWh", vals, 5)
        kwh = bill.kwh if kwh_str == NO_CHANGE else Decimal(kwh_str)

        net_str = add_arg(args, "Net", vals, 6)
        net = bill.net if net_str == NO_CHANGE else Decimal(net_str)

        vat_str = add_arg(args, "Vat", vals, 7)
        vat = bill.vat if vat_str == NO_CHANGE else Decimal(vat_str)

        gross_str = add_arg(args, "Gross", vals, 8)
        gross = bill.gross if gross_str == NO_CHANGE else Decimal(gross_str)

        bill_type_code = add_arg(args, "Bill Type", vals, 9)
        if bill_type_code == NO_CHANGE:
            bill_type = bill.bill_type
        else:
            bill_type = BillType.get_by_code(sess, bill_type_code)

        breakdown = add_arg(args, "Breakdown", vals, 10)
        if breakdown == NO_CHANGE:
            breakdown = bill.breakdown

        bill.update(
            "",
            reference,
            issue_date,
            start_date,
            finish_date,
            kwh,
            net,
            vat,
            gross,
            bill_type,
            breakdown,
        )

    elif action == "delete":
        bill_id_str = add_arg(args, "Bill Chellow Id", vals, 0)
        bill_id = int(bill_id_str)
        bill = Bill.get_by_id(sess, bill_id)
        bill.delete(sess)


def general_import_mop_bill_element(sess, action, vals, args):
    vals = _truncate_vals(vals)
    if action == "insert":
        contract_name = add_arg(args, "Contract Name", vals, 0)
        contract = Contract.get_mop_by_name(sess, contract_name)
        batch_reference = add_arg(args, "Batch Reference", vals, 1)
        batch = contract.get_batch(sess, batch_reference)
        bill_reference = add_arg(args, "Bill Reference", vals, 2)
        bill = batch.get_bill_by_reference(sess, bill_reference)
        name = add_arg(args, "Name", vals, 3)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        net_str = add_arg(args, "Net", vals, 6)
        net = Decimal(net_str)
        breakdown_str = add_arg(args, "Breakdown", vals, 7)
        breakdown = _parse_breakdown(breakdown_str)
        bill.insert_element(
            sess,
            name,
            start_date,
            finish_date,
            net,
            breakdown,
        )


def general_import_dc_bill(sess, action, vals, args):
    vals = _truncate_vals(vals)
    if action == "insert":
        contract_name = add_arg(args, "Contract Name", vals, 0)
        contract = Contract.get_dc_by_name(sess, contract_name)

        batch_reference = add_arg(args, "Batch Reference", vals, 1)

        batch = contract.get_batch(sess, batch_reference)
        mpan_core_str = add_arg(args, "Mpan Core", vals, 2)
        mpan_core = parse_mpan_core(mpan_core_str)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        issue_date_str = add_arg(args, "Issue Date", vals, 3)
        issue_date = parse_hh_start(issue_date_str)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        net_str = add_arg(args, "Net", vals, 6)
        net = Decimal(net_str)
        vat_str = add_arg(args, "Vat", vals, 7)
        vat = Decimal(vat_str)
        gross_str = add_arg(args, "Gross", vals, 8)
        gross = Decimal(gross_str)
        reference = add_arg(args, "Reference", vals, 9)
        type_code = add_arg(args, "Type", vals, 10)
        typ = BillType.get_by_code(sess, type_code)
        breakdown_str = add_arg(args, "Breakdown", vals, 11)
        breakdown = _parse_breakdown(breakdown_str)

        bill = batch.insert_bill(
            sess,
            "",
            reference,
            issue_date,
            start_date,
            finish_date,
            0,
            net,
            vat,
            gross,
            typ,
            breakdown,
            supply,
        )

    elif action == "update":
        bill_id_str = add_arg(args, "Bill Id", vals, 0)
        bill_id = int(bill_id_str)
        bill = Bill.get_by_id(bill_id)

        reference = add_arg(args, "Reference", vals, 1)
        if reference == NO_CHANGE:
            reference = bill.getReference()

        issue_date_str = add_arg(args, "Issue Date", vals, 2)
        issue_date = None

        if issue_date_str == NO_CHANGE:
            issue_date = bill.issue_date
        else:
            issue_date = parse_hh_start(issue_date_str)

        start_date_str = add_arg(args, "Start Date", vals, 3)
        if start_date_str == NO_CHANGE:
            start_date = bill.start_date
        else:
            start_date = parse_hh_start(start_date_str)

        finish_date_str = add_arg(args, "Finish Date", vals, 4)
        if finish_date_str == NO_CHANGE:
            finish_date = bill.finish_date
        else:
            finish_date = parse_hh_start(finish_date_str)

        kwh_str = add_arg(args, "kWh", vals, 5)
        kwh = bill.kwh if kwh_str == NO_CHANGE else Decimal(kwh_str)

        net_str = add_arg(args, "Net", vals, 6)
        net = bill.net if net_str == NO_CHANGE else Decimal(net_str)

        vat_str = add_arg(args, "Vat", vals, 7)
        vat = bill.vat if vat_str == NO_CHANGE else Decimal(vat_str)

        gross_str = add_arg(args, "Gross", vals, 8)
        gross = bill.gross if gross_str == NO_CHANGE else Decimal(gross_str)

        bill_type_code = add_arg(args, "Bill Type", vals, 9)
        if bill_type_code == NO_CHANGE:
            bill_type = bill.bill_type
        else:
            bill_type = BillType.get_by_code(sess, bill_type_code)

        breakdown = add_arg(args, "Breakdown", vals, 10)
        if breakdown == NO_CHANGE:
            breakdown = bill.breakdown

        bill.update(
            "",
            reference,
            issue_date,
            start_date,
            finish_date,
            kwh,
            net,
            vat,
            gross,
            bill_type,
            breakdown,
        )

    elif action == "delete":
        bill_id_str = add_arg(args, "Bill Chellow Id", vals, 0)
        bill_id = int(bill_id_str)
        bill = Bill.get_by_id(sess, bill_id)
        bill.delete(sess)


def general_import_dc_bill_element(sess, action, vals, args):
    vals = _truncate_vals(vals)
    if action == "insert":
        contract_name = add_arg(args, "Contract Name", vals, 0)
        contract = Contract.get_dc_by_name(sess, contract_name)
        batch_reference = add_arg(args, "Batch Reference", vals, 1)
        batch = contract.get_batch(sess, batch_reference)
        bill_reference = add_arg(args, "Bill Reference", vals, 2)
        bill = batch.get_bill_by_reference(sess, bill_reference)
        name = add_arg(args, "Name", vals, 3)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        net_str = add_arg(args, "Net", vals, 6)
        net = Decimal(net_str)
        breakdown_str = add_arg(args, "Breakdown", vals, 7)
        breakdown = _parse_breakdown(breakdown_str)
        bill.insert_element(
            sess,
            name,
            start_date,
            finish_date,
            net,
            breakdown,
        )


def general_import_supplier_bill(sess, action, vals, args):
    vals = _truncate_vals(vals)
    if action == "insert":
        contract_name = add_arg(args, "Contract Name", vals, 0)
        contract = Contract.get_supplier_by_name(sess, contract_name)
        batch_reference = add_arg(args, "Batch Reference", vals, 1)

        batch = contract.get_batch(sess, batch_reference)
        mpan_core_str = add_arg(args, "Mpan Core", vals, 2)
        mpan_core = parse_mpan_core(mpan_core_str)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        issue_date_str = add_arg(args, "Issue Date", vals, 3)
        issue_date = parse_hh_start(issue_date_str)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        net_str = add_arg(args, "Net", vals, 6)
        net = Decimal(net_str)
        vat_str = add_arg(args, "Vat", vals, 7)
        vat = Decimal(vat_str)
        gross_str = add_arg(args, "Gross", vals, 8)
        gross = Decimal(gross_str)
        account = add_arg(args, "Account Reference", vals, 9)
        reference = add_arg(args, "Reference", vals, 10)
        type_code = add_arg(args, "Type", vals, 11)
        typ = BillType.get_by_code(sess, type_code)
        breakdown_str = add_arg(args, "Breakdown", vals, 12)
        breakdown = _parse_breakdown(breakdown_str)

        kwh_str = add_arg(args, "kWh", vals, 13)
        kwh = Decimal(kwh_str)

        bill = batch.insert_bill(
            sess,
            account,
            reference,
            issue_date,
            start_date,
            finish_date,
            kwh,
            net,
            vat,
            gross,
            typ,
            breakdown,
            supply,
        )

        i = 14
        while i < len(vals):
            typ = add_arg(args, "read", vals, i)
            if typ == "read":
                msn = add_arg(args, "Meter Serial Number", vals, i + 1)
                mpan_str = add_arg(args, "MPAN", vals, i + 2)
                coefficient_str = add_arg(args, "Coefficient", vals, i + 3)
                coefficient = Decimal(coefficient_str)
                units = add_arg(args, "Units", vals, i + 4)
                tpr_code = add_arg(args, "TPR", vals, i + 5)
                if len(tpr_code) > 0:
                    tpr = Tpr.get_by_code(sess, tpr_code)
                else:
                    tpr = None

                prev_date_str = add_arg(args, "Previous Date", vals, i + 6)
                prev_date = parse_hh_start(prev_date_str)
                prev_value_str = add_arg(args, "Previous Value", vals, i + 7)
                prev_value = Decimal(prev_value_str)

                prev_type_str = add_arg(args, "Previous Type", vals, i + 8)
                prev_type = ReadType.get_by_code(sess, prev_type_str)

                pres_date_str = add_arg(args, "Present Date", vals, i + 9)
                pres_date = parse_hh_start(pres_date_str)
                pres_value_str = add_arg(args, "Present Value", vals, i + 10)
                pres_value = Decimal(pres_value_str)

                pres_type_str = add_arg(args, "Present Type", vals, i + 11)
                pres_type = ReadType.get_by_code(sess, pres_type_str)
                bill.insert_read(
                    sess,
                    tpr,
                    coefficient,
                    units,
                    msn,
                    mpan_str,
                    prev_date,
                    prev_value,
                    prev_type,
                    pres_date,
                    pres_value,
                    pres_type,
                )
                i += 12
            else:
                raise BadRequest(
                    f"The type must be 'read', but found {typ} at column {i}."
                )

    elif action == "update":
        bill_id_str = add_arg(args, "Bill Id", vals, 0)
        bill_id = int(bill_id_str)
        bill = Bill.get_by_id(bill_id)

        account = add_arg(args, "Account Reference", vals, 1)
        if account == NO_CHANGE:
            account = bill.getAccount()

        reference = add_arg(args, "Reference", vals, 2)
        if reference == NO_CHANGE:
            reference = bill.getReference()

        issue_date_str = add_arg(args, "Issue Date", vals, 3)
        issue_date = None

        if issue_date_str == NO_CHANGE:
            issue_date = bill.issue_date
        else:
            issue_date = parse_hh_start(issue_date_str)

        start_date_str = add_arg(args, "Start Date", vals, 4)
        if start_date_str == NO_CHANGE:
            start_date = bill.start_date
        else:
            start_date = parse_hh_start(start_date_str)

        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        if finish_date_str == NO_CHANGE:
            finish_date = bill.finish_date
        else:
            finish_date = parse_hh_start(finish_date_str)

        kwh_str = add_arg(args, "kWh", vals, 6)
        kwh = bill.kwh if kwh_str == NO_CHANGE else Decimal(kwh_str)

        net_str = add_arg(args, "Net", vals, 7)
        net = bill.net if net_str == NO_CHANGE else Decimal(net_str)

        vat_str = add_arg(args, "Vat", vals, 8)
        vat = bill.vat if vat_str == NO_CHANGE else Decimal(vat_str)

        gross_str = add_arg(args, "Gross", vals, 9)
        gross = bill.gross if gross_str == NO_CHANGE else Decimal(gross_str)

        bill_type_code = add_arg(args, "Bill Type", vals, 10)
        if bill_type_code == NO_CHANGE:
            bill_type = bill.bill_type
        else:
            bill_type = BillType.get_by_code(sess, bill_type_code)

        breakdown = add_arg(args, "Breakdown", vals, 11)
        if breakdown == NO_CHANGE:
            breakdown = bill.breakdown

        bill.update(
            account,
            reference,
            issue_date,
            start_date,
            finish_date,
            kwh,
            net,
            vat,
            gross,
            bill_type,
            breakdown,
        )

    elif action == "delete":
        bill_id_str = add_arg(args, "Bill Chellow Id", vals, 0)
        bill_id = int(bill_id_str)
        bill = Bill.get_by_id(sess, bill_id)
        bill.delete(sess)


def general_import_supplier_bill_element(sess, action, vals, args):
    vals = _truncate_vals(vals)
    if action == "insert":
        contract_name = add_arg(args, "Contract Name", vals, 0)
        contract = Contract.get_supplier_by_name(sess, contract_name)
        batch_reference = add_arg(args, "Batch Reference", vals, 1)
        batch = contract.get_batch(sess, batch_reference)
        bill_reference = add_arg(args, "Bill Reference", vals, 2)
        bill = batch.get_bill_by_reference(sess, bill_reference)
        name = add_arg(args, "Name", vals, 3)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        net_str = add_arg(args, "Net", vals, 6)
        net = Decimal(net_str)
        breakdown_str = add_arg(args, "Breakdown", vals, 7)
        breakdown = _parse_breakdown(breakdown_str)
        bill.insert_element(
            sess,
            name,
            start_date,
            finish_date,
            net,
            breakdown,
        )


def general_import_g_bill(sess, action, vals, args):
    if action == "insert":
        contract_name = add_arg(args, "Supplier Contract Name", vals, 0)

        g_contract = GContract.get_supplier_by_name(sess, contract_name)

        batch_reference = add_arg(args, "Batch Reference", vals, 1)

        g_batch = g_contract.get_g_batch_by_reference(sess, batch_reference)

        mprn = add_arg(args, "MPRN", vals, 2)
        g_supply = GSupply.get_by_mprn(sess, mprn)

        issue_date_str = add_arg(args, "Issue Date", vals, 3)
        issue_date = parse_hh_start(issue_date_str)
        start_date_str = add_arg(args, "Start Date", vals, 4)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 5)
        finish_date = parse_hh_start(finish_date_str)
        net_gbp_str = add_arg(args, "Net GBP", vals, 6)
        net_gbp = Decimal(net_gbp_str)
        vat_gbp_str = add_arg(args, "Vat GBP", vals, 7)
        vat_gbp = Decimal(vat_gbp_str)
        gross_gbp_str = add_arg(args, "Gross GBP", vals, 8)
        gross_gbp = Decimal(gross_gbp_str)
        account = add_arg(args, "Account Reference", vals, 9)
        reference = add_arg(args, "Reference", vals, 10)
        bill_type_code = add_arg(args, "Type", vals, 11)
        bill_type = BillType.get_by_code(sess, bill_type_code)
        breakdown_str = add_arg(args, "Breakdown", vals, 12)
        breakdown = _parse_breakdown(breakdown_str)

        kwh_str = add_arg(args, "kWh", vals, 13)
        kwh = Decimal(kwh_str)

        g_bill = g_batch.insert_g_bill(
            sess,
            g_supply,
            bill_type,
            reference,
            account,
            issue_date,
            start_date,
            finish_date,
            kwh,
            net_gbp,
            vat_gbp,
            gross_gbp,
            "",
            breakdown,
        )

        for i in range(14, len(vals), 10):
            msn = add_arg(args, "Meter Serial Number", vals, i)
            g_unit_code = add_arg(args, "Unit", vals, i + 1)
            g_unit = GUnit.get_by_code(sess, g_unit_code)
            correction_factor_str = add_arg(args, "Correction Factor", vals, i + 2)
            correction_factor = Decimal(correction_factor_str)
            calorific_value_str = add_arg(args, "Calorific Value", vals, i + 3)
            calorific_value = Decimal(calorific_value_str)

            prev_date_str = add_arg(args, "Previous Date", vals, i + 4)
            prev_date = parse_hh_start(prev_date_str)
            prev_value_str = add_arg(args, "Previous Value", vals, i + 5)
            prev_value = Decimal(prev_value_str)

            prev_type_str = add_arg(args, "Previous Type", vals, i + 6)
            prev_type = GReadType.get_by_code(sess, prev_type_str)

            pres_date_str = add_arg(args, "Present Date", vals, i + 7)
            pres_date = parse_hh_start(pres_date_str)
            pres_value_str = add_arg(args, "Present Value", vals, i + 8)
            pres_value = Decimal(pres_value_str)

            pres_type_str = add_arg(args, "Present Type", vals, i + 9)
            pres_type = GReadType.get_by_code(sess, pres_type_str)

            g_bill.insert_g_read(
                sess,
                msn,
                g_unit,
                correction_factor,
                calorific_value,
                prev_value,
                prev_date,
                prev_type,
                pres_value,
                pres_date,
                pres_type,
            )


def general_import_g_register_read(sess, action, vals, args):
    if action == "insert":
        contract_name = add_arg(args, "Supplier Contract Name", vals, 0)

        g_contract = GContract.get_supplier_by_name(sess, contract_name)

        batch_reference = add_arg(args, "Batch Reference", vals, 1)

        g_batch = g_contract.get_g_batch_by_reference(sess, batch_reference)

        mprn = add_arg(args, "MPRN", vals, 2)
        g_supply = GSupply.get_by_mprn(sess, mprn)

        bill_start_date_str = add_arg(args, "Bill Start Date", vals, 3)
        bill_start_date = parse_hh_start(bill_start_date_str)

        g_bill = sess.scalars(
            select(GBill).where(
                GBill.g_batch == g_batch,
                GBill.g_supply == g_supply,
                GBill.start_date == bill_start_date,
            )
        ).first()

        if g_bill is None:
            raise BadRequest(
                f"Can't find a bill in batch {batch_reference} in contract "
                f"{contract_name} with MPRN {mprn} starting at "
                f"{hh_format(bill_start_date)}"
            )

        for i in range(4, len(vals), 10):
            msn = add_arg(args, "Meter Serial Number", vals, i)
            g_unit_code = add_arg(args, "Unit", vals, i + 1)
            g_unit = GUnit.get_by_code(sess, g_unit_code)
            correction_factor_str = add_arg(args, "Correction Factor", vals, i + 2)
            correction_factor = Decimal(correction_factor_str)
            calorific_value_str = add_arg(args, "Calorific Value", vals, i + 3)
            calorific_value = Decimal(calorific_value_str)

            prev_date_str = add_arg(args, "Previous Date", vals, i + 4)
            prev_date = parse_hh_start(prev_date_str)
            prev_value_str = add_arg(args, "Previous Value", vals, i + 5)
            prev_value = Decimal(prev_value_str)

            prev_type_str = add_arg(args, "Previous Type", vals, i + 6)
            prev_type = GReadType.get_by_code(sess, prev_type_str)

            pres_date_str = add_arg(args, "Present Date", vals, i + 7)
            pres_date = parse_hh_start(pres_date_str)
            pres_value_str = add_arg(args, "Present Value", vals, i + 8)
            pres_value = Decimal(pres_value_str)

            pres_type_str = add_arg(args, "Present Type", vals, i + 9)
            pres_type = GReadType.get_by_code(sess, pres_type_str)

            g_bill.insert_g_read(
                sess,
                msn,
                g_unit,
                correction_factor,
                calorific_value,
                prev_value,
                prev_date,
                prev_type,
                pres_value,
                pres_date,
                pres_type,
            )


def general_import_register_read(sess, action, vals, args):
    if action == "insert":
        pass
    elif action == "update":
        read_id_str = add_arg(args, "Chellow Id", vals, 0)
        read = RegisterRead.get_by_id(int(read_id_str))

        tpr_code = add_arg(args, "TPR", vals, 1)
        if tpr_code == NO_CHANGE:
            tpr = read.tpr
        else:
            tpr = Tpr.get_by_code(sess, tpr_code)

        coefficient_str = add_arg(args, "Coefficient", vals, 2)
        if coefficient_str == NO_CHANGE:
            coefficient = read.coefficient
        else:
            coefficient = Decimal(coefficient_str)

        units = add_arg(args, "Units", vals, 3)

        msn = add_arg(args, "Meter Serial Number", vals, 4)
        if msn == NO_CHANGE:
            msn = read.msn

        mpan_str = add_arg(args, "MPAN", vals, 5)
        if mpan_str == NO_CHANGE:
            mpan_str = read.mpan_str

        prev_date_str = add_arg(args, "Previous Date", vals, 6)
        if prev_date_str == NO_CHANGE:
            prev_date = read.prev_date
        else:
            prev_date = parse_hh_start(prev_date_str)

        prev_value_str = add_arg(args, "Previous Value", vals, 7)
        if prev_value_str == NO_CHANGE:
            prev_value = read.prev_value
        else:
            prev_value = parse_hh_start(prev_value_str)

        prev_type_code = add_arg(args, "Previous Type", vals, 8)
        if prev_type_code == NO_CHANGE:
            prev_type = read.prev_type
        else:
            prev_type = ReadType.get_by_code(sess, prev_type_code)

        pres_date_str = add_arg(args, "Present Date", vals, 9)
        if pres_date_str == NO_CHANGE:
            pres_date = read.pres_date
        else:
            pres_date = parse_hh_start(pres_date_str)

        pres_value_str = add_arg(args, "Present Value", vals, 10)
        if pres_value_str == NO_CHANGE:
            pres_value = read.pres_value
        else:
            pres_value = parse_hh_start(pres_value_str)

        pres_type_code = add_arg(args, "Present Type", vals, 11)
        if pres_type_code == NO_CHANGE:
            pres_type = read.pres_type
        else:
            pres_type = ReadType.get_by_code(sess, pres_type_code)

        read.update(
            tpr,
            coefficient,
            units,
            msn,
            mpan_str,
            prev_date,
            prev_value,
            prev_type,
            pres_date,
            pres_value,
            pres_type,
        )


def general_import_supply(sess, action, vals, args):
    if action == "insert":
        site_code = add_arg(args, "Site Code", vals, 0)
        site = Site.get_by_code(sess, site_code)
        source_code = add_arg(args, "Source Code", vals, 1)
        source = Source.get_by_code(sess, source_code)
        gen_type_code = add_arg(args, "Generator Type", vals, 2)
        gen_type = None
        if source.code == "gen" or source.code == "gen-net":
            gen_type = GeneratorType.get_by_code(sess, gen_type_code)

        supply_name = add_arg(args, "Supply Name", vals, 3)
        gsp_group_code = add_arg(args, "Gsp Group", vals, 4)
        gsp_group = GspGroup.get_by_code(sess, gsp_group_code)
        start_date_str = add_arg(args, "Start Date", vals, 5)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 6)
        finish_date = parse_hh_start(finish_date_str)
        mop_contract_name = add_arg(args, "MOP Contract", vals, 7)
        mop_contract = Contract.get_mop_by_name(sess, mop_contract_name)
        dc_contract_name = add_arg(args, "DC Contract", vals, 8)
        dc_contract = Contract.get_dc_by_name(sess, dc_contract_name)
        msn = add_arg(args, "Meter Serial Number", vals, 9)
        dno_code = add_arg(args, "DNO Code", vals, 10)
        dno = Party.get_dno_by_code(sess, dno_code, start_date)
        pc_code = add_arg(args, "Profile Class", vals, 11)
        pc = Pc.get_by_code(sess, parse_pc_code(pc_code))
        mtc_code = add_arg(args, "Meter Timeswitch Class", vals, 12)
        cop_code = add_arg(args, "CoP", vals, 13)
        cop = Cop.get_by_code(sess, cop_code)
        comm_code = add_arg(args, "Comms Type", vals, 14)
        comm = Comm.get_by_code(sess, comm_code)
        ssc_code_str = add_arg(args, "Standard Settlement Configuration", vals, 15)
        if len(ssc_code_str) > 0:
            ssc = Ssc.get_by_code(sess, ssc_code_str, start_date)
            ssc_code = ssc.code
        else:
            ssc_code = None
        energisation_status_code = add_arg(args, "Energisation Status", vals, 16)
        energisation_status = EnergisationStatus.get_by_code(
            sess, energisation_status_code
        )
        dtc_meter_type_code = add_arg(args, "DTC Meter Type", vals, 17)
        dtc_meter_type = DtcMeterType.get_by_code(sess, dtc_meter_type_code)
        imp_mpan_core = add_arg(args, "Import MPAN Core", vals, 18)
        if len(imp_mpan_core) == 0:
            imp_mpan_core = None
        else:
            imp_mpan_core = parse_mpan_core(imp_mpan_core)

        if imp_mpan_core is None:
            imp_llfc_code = None
            imp_supplier_contract = None
            imp_supplier_account = None
            imp_sc = None
        else:
            imp_llfc_code = add_arg(args, "Import LLFC", vals, 19)
            imp_sc_str = add_arg(args, "Import Agreed Supply Capacity", vals, 20)
            try:
                imp_sc = int(imp_sc_str)
            except ValueError as e:
                raise BadRequest(f"The import supply capacity must be an integer. {e}")

            imp_supplier_contract_name = add_arg(
                args, "Import Supplier Contract", vals, 21
            )
            imp_supplier_account = add_arg(args, "Import Supplier Account", vals, 22)
            imp_supplier_contract = Contract.get_supplier_by_name(
                sess, imp_supplier_contract_name
            )

        exp_supplier_contract = exp_sc = exp_llfc_code = exp_mpan_core = None
        exp_supplier_account = None
        if len(vals) > 23:
            exp_mpan_core = add_arg(args, "Export MPAN Core", vals, 23)
            if len(exp_mpan_core) == 0:
                exp_mpan_core = None
            else:
                exp_mpan_core = parse_mpan_core(exp_mpan_core)

            if exp_mpan_core is not None:
                exp_llfc_code = add_arg(args, "Export LLFC", vals, 24)
                exp_sc_str = add_arg(args, "Export Agreed Supply Capacity", vals, 25)
                try:
                    exp_sc = int(exp_sc_str)
                except ValueError as e:
                    raise BadRequest(
                        f"The export agreed supply capacity must be an integer. {e}"
                    )

                exp_supplier_contract_name = add_arg(
                    args, "Export Supplier Contract", vals, 26
                )
                exp_supplier_contract = Contract.get_supplier_by_name(
                    sess, exp_supplier_contract_name
                )
                exp_supplier_account = add_arg(
                    args, "Export Supplier Account", vals, 27
                )

        supply = site.insert_e_supply(
            sess,
            source,
            gen_type,
            supply_name,
            start_date,
            finish_date,
            gsp_group,
            mop_contract,
            dc_contract,
            msn,
            dno,
            pc,
            mtc_code,
            cop,
            comm,
            ssc_code,
            energisation_status,
            dtc_meter_type,
            imp_mpan_core,
            imp_llfc_code,
            imp_supplier_contract,
            imp_supplier_account,
            imp_sc,
            exp_mpan_core,
            exp_llfc_code,
            exp_supplier_contract,
            exp_supplier_account,
            exp_sc,
        )
        sess.flush()

    elif action == "update":
        mpan_core = add_arg(args, "MPAN Core", vals, 0)
        source_code = add_arg(args, "Source Code", vals, 1)
        gen_type_code = add_arg(args, "Generator Type", vals, 2)
        gsp_group_code = add_arg(args, "GSP Group", vals, 3)
        supply_name = add_arg(args, "Supply Name", vals, 4)
        supply = Supply.get_by_mpan_core(sess, mpan_core)

        supply_name = supply.name if supply_name == NO_CHANGE else supply_name
        if source_code == NO_CHANGE:
            source = Source.get_by_code(sess, source_code)
        else:
            source = supply.source

        if gen_type_code == NO_CHANGE:
            gen_type = supply.generator_type
        else:
            gen_type = GeneratorType.get_by_code(sess, gen_type_code)

        if gsp_group_code == NO_CHANGE:
            gsp_group = supply.gsp_group
        else:
            gsp_group = GspGroup.get_by_code(sess, gsp_group_code)

        supply.update(supply_name, source, gen_type, gsp_group)

    elif action == "delete":
        mpan_core = add_arg(args, "MPAN Core", vals, 0)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        supply.delete()


def general_import_llfc(sess, action, vals, args):
    if action == "insert":
        dno_code = add_arg(args, "dno", vals, 0)
        llfc_code = add_arg(args, "llfc", vals, 1)
        llfc_description = add_arg(args, "llfc_description", vals, 2)
        vl_code = add_arg(args, "voltage_level", vals, 3)
        vl = VoltageLevel.get_by_code(sess, vl_code.upper())
        is_substation_str = add_arg(args, "is_substation", vals, 4)
        is_substation = parse_bool(is_substation_str)
        is_import_str = add_arg(args, "is_import", vals, 5)
        is_import = parse_bool(is_import_str)
        valid_from_str = add_arg(args, "valid_from", vals, 6)
        valid_from = parse_hh_start(valid_from_str)
        valid_to_str = add_arg(args, "valid_to", vals, 7)
        valid_to = parse_hh_start(valid_to_str)

        dno = Party.get_dno_by_code(sess, dno_code, valid_from)
        existing_llfc = dno.find_llfc_by_code(sess, llfc_code, valid_from)

        if existing_llfc is None:
            llfc = Llfc(
                dno,
                llfc_code,
                llfc_description,
                vl,
                is_substation,
                is_import,
                valid_from,
                valid_to,
            )
            sess.add(llfc)
            sess.flush()
        else:
            raise BadRequest(
                "There's already a LLFC with this DNO and code for this " "period."
            )

    elif action == "update":
        dno_code = add_arg(args, "dno", vals, 0)
        llfc_code = add_arg(args, "llfc", vals, 1)
        valid_from_str = add_arg(args, "valid_from", vals, 2)
        valid_from = parse_hh_start(valid_from_str)

        dno = Party.get_dno_by_code(sess, dno_code, valid_from)
        llfc = (
            sess.query(Llfc)
            .filter(
                Llfc.dno == dno, Llfc.code == llfc_code, Llfc.valid_from == valid_from
            )
            .first()
        )
        if llfc is None:
            raise BadRequest(
                f"Can't find an LLFC for the DNO {dno_code} and 'valid from' "
                f"date {hh_format(valid_from)}."
            )

        llfc_description_str = add_arg(args, "llfc_description", vals, 3)
        if llfc_description_str == NO_CHANGE:
            llfc_description = llfc.description
        else:
            llfc_description = llfc_description_str

        vl_code_str = add_arg(args, "voltage_level", vals, 4)
        if vl_code_str == NO_CHANGE:
            vl = llfc.voltage_level
        else:
            vl = VoltageLevel.get_by_code(sess, vl_code_str.upper())

        is_substation_str = add_arg(args, "is_substation", vals, 5)
        if is_substation_str == NO_CHANGE:
            is_substation = llfc.is_substation
        else:
            is_substation = parse_bool(is_substation_str)

        is_import_str = add_arg(args, "is_import", vals, 6)
        if is_import_str == NO_CHANGE:
            is_import = llfc.is_import
        else:
            is_import = parse_bool(is_import_str)

        valid_to_str = add_arg(args, "valid_to", vals, 7)
        if valid_to_str == NO_CHANGE:
            valid_to = llfc.valid_to
        else:
            valid_to = parse_hh_start(valid_to_str)

        llfc.update(
            llfc_description, vl, is_substation, is_import, llfc.valid_from, valid_to
        )
        sess.flush()

    elif action == "delete":
        dno_code = add_arg(args, "dno_code", vals, 0)
        dno = Party.get_dno_by_code(sess, dno_code)
        llfc_code = add_arg(args, "llfc", vals, 1)
        date_str = add_arg(args, "date", vals, 2)
        date = parse_hh_start(date_str)

        llfc = dno.get_llfc_by_code(sess, llfc_code, date)
        sess.delete(llfc)
        sess.flush()
    else:
        raise BadRequest("Action not recognized.")


def general_import_site_era(sess, action, vals, args):
    site_code = add_arg(args, "site_code", vals, 0)
    site = Site.get_by_code(sess, site_code)
    mpan_core = add_arg(args, "mpan_core", vals, 1)
    start_date_str = add_arg(args, "generation_start_date", vals, 2)
    start_date = datetime.datetime(start_date_str)
    era = Era.get_by_core_date(mpan_core, start_date)
    if action == "insert":
        is_location_str = add_arg(args, "is_location", vals, 3)
        is_location = bool(is_location_str)
        era.attach_site(site, is_location)
        sess.flush()


def general_import_channel(sess, action, vals, args):
    mpan_core_raw = add_arg(args, "MPAN Core", vals, 0)
    mpan_core = parse_mpan_core(mpan_core_raw)
    supply = Supply.find_by_mpan_core(sess, mpan_core)
    dt_raw = add_arg(args, "Date", vals, 1)
    dt = parse_hh_start(dt_raw)
    era = supply.find_era_at(sess, dt)
    import_related_str = add_arg(args, "Import Related?", vals, 2)
    import_related = parse_bool(import_related_str)
    channel_type_raw = add_arg(args, "Channel Type", vals, 3)
    channel_type = parse_channel_type(channel_type_raw)

    if action == "insert":
        era.insert_channel(sess, import_related, channel_type)
    elif action == "delete":
        era.delete_channel(sess, import_related, channel_type)


def general_import_user(sess, action, vals, args):
    if action == "insert":
        email_address = add_arg(args, "email_address", vals, 0)
        password = add_arg(args, "password", vals, 1)
        digest = add_arg(args, "password_digest", vals, 2)
        user_role_code = add_arg(args, "user_role_code", vals, 3)
        user_role = UserRole.get_by_code(sess, user_role_code)
        participant_code = add_arg(args, "participant_code", vals, 4)
        party = None
        if len(participant_code.strip()) > 0:
            market_role_code = add_arg(args, "market_role_code", vals, 5)
        party = Party.get_by_participant_role(participant_code, market_role_code)
        if len(password) == 0:
            if len(digest) == 0:
                raise BadRequest("The password and digest fields can't both be blank.")
        elif len(digest) > 0:
            raise BadRequest("The password and digest fields can't both be filled.")
        else:
            digest = User.digest(password)

        User.insert(email_address, user_role, party, digest)
    elif action == "update":
        pass


def general_import_site(sess, action, vals, args):
    code = add_arg(args, "site_code", vals, 0)

    if action == "insert":
        name = add_arg(args, "site_name", vals, 1)
        Site.insert(sess, code, name)
    else:
        site = Site.get_by_code(sess, code)

        if action == "delete":
            site.delete()
        elif action == "update":
            new_code = add_arg(args, "new_site_code", vals, 1)
            if new_code == NO_CHANGE:
                new_code = site.code
            name = add_arg(args, "new_site_name", vals, 2)
            if name == NO_CHANGE:
                name = site.name
            site.update(new_code, name)


def general_import_batch(sess, action, vals, args):
    if action == "insert":
        role_name = add_arg(args, "Role Name", vals, 0).lower()
        contract_name = add_arg(args, "Contract Name", vals, 1)

        if role_name == "dc":
            contract = Contract.get_dc_by_name(sess, contract_name)
        elif role_name == "supplier":
            contract = Contract.get_supplier_by_name(sess, contract_name)
        elif role_name == "mop":
            contract = Contract.get_mop_by_name(sess, contract_name)
        else:
            raise BadRequest("The role name must be one of dc, supplier or mop.")

        reference = add_arg(args, "Reference", vals, 2)
        description = add_arg(args, "Description", vals, 3)
        contract.insert_batch(sess, reference, description)
    elif action == "update":
        role_name = add_arg(args, "Role Name", vals, 0).lower()
        contract_name = add_arg(args, "Contract Name", vals, 1)

        if role_name == "dc":
            contract = Contract.get_dc_by_name(sess, contract_name)
        elif role_name == "supplier":
            contract = Contract.get_supplier_by_name(sess, contract_name)
        elif role_name == "mop":
            contract = Contract.get_mop_by_name(sess, contract_name)
        else:
            raise BadRequest("The role name must be one of dc, supplier or mop.")

        old_reference = add_arg(args, "Old Reference", vals, 2)
        batch = contract.get_batch(sess, sess, old_reference)
        new_reference = add_arg(args, "New Reference", vals, 3)
        description = add_arg(args, "Description", vals, 4)
        batch.update(sess, new_reference, description)


def general_import_site_snag_ignore(sess, action, vals, args):
    if action == "insert":
        site_code = add_arg(args, "Site Code", vals, 0)
        site = Site.get_by_code(sess, site_code)
        description = add_arg(args, "Snag Description", vals, 1)
        start_date_str = add_arg(args, "Start Date", vals, 2)
        start_date = parse_hh_start(start_date_str)
        finish_date_str = add_arg(args, "Finish Date", vals, 3)
        finish_date = parse_hh_start(finish_date_str)

        for snag in sess.query(Snag).filter(
            Snag.site_id == site.id,
            Snag.description == description,
            Snag.is_ignored != false(),
            Snag.start_date <= finish_date,
            or_(Snag.finish_date == null(), Snag.finish_date >= start_date),
        ):
            snag.set_is_ignored(True)

    elif action == "update":
        raise BadRequest("The 'update' action isn't supported for site snags.")


def general_import_channel_snag_ignore(sess, action, vals, args):
    _channel_snag_update(sess, action, vals, args, True)


def general_import_channel_snag_unignore(sess, action, vals, args):
    _channel_snag_update(sess, action, vals, args, False)


def _channel_snag_update(sess, action, vals, args, ignore):
    if action == "insert":
        mpan_core_str = add_arg(args, "MPAN Core", vals, 0)
        mpan_core = parse_mpan_core(mpan_core_str)
        supply = Supply.get_by_mpan_core(sess, mpan_core)
        imp_related_str = add_arg(args, "Import Related?", vals, 1)
        imp_related = parse_bool(imp_related_str)
        channel_type_str = add_arg(args, "Channel Type", vals, 2)
        channel_type = parse_channel_type(channel_type_str)
        description = add_arg(args, "Snag Description", vals, 3)
        start_str = add_arg(args, "From", vals, 4)
        start_date = parse_hh_start(start_str)
        finish_str = add_arg(args, "To", vals, 5)
        finish_date = parse_hh_start(finish_str)

        for era in supply.find_eras(sess, start_date, finish_date):
            channel = (
                sess.query(Channel)
                .filter(
                    Channel.era == era,
                    Channel.imp_related == imp_related,
                    Channel.channel_type == channel_type,
                )
                .first()
            )
            if channel is not None:
                snag_query = sess.query(Snag).filter(
                    Snag.channel == channel,
                    Snag.is_ignored == (not ignore),
                    Snag.description == description,
                    or_(Snag.finish_date == null(), Snag.finish_date >= start_date),
                )

            if finish_date is not None:
                snag_query = snag_query.filter(Snag.start_date <= finish_date)

            for snag in snag_query:
                snag.set_is_ignored(ignore)

    elif action == "update":
        raise BadRequest("The action 'update' isn't supported for channel snags.")


PREFIX = "general_import_"

typ_funcs = {}
for k in tuple(globals().keys()):
    if k.startswith(PREFIX):
        typ_funcs[k[len(PREFIX) :]] = globals()[k]


def _process_line(sess, hh_data, line, args):

    action = add_arg(args, "action", line, 0).lower()
    if action not in ALLOWED_ACTIONS:
        raise BadRequest(f"The 'action' field must be one of {ALLOWED_ACTIONS}")
    typ = add_arg(args, "type", line, 1).lower()
    vals = line[2:]
    if typ == "hh_datum":
        if action == "insert":
            hh_data.append(
                {
                    "mpan_core": parse_mpan_core(add_arg(args, "MPAN Core", vals, 0)),
                    "start_date": parse_hh_start(add_arg(args, "Start Date", vals, 1)),
                    "channel_type": parse_channel_type(
                        add_arg(args, "Channel Type", vals, 2)
                    ),
                    "value": Decimal(add_arg(args, "Value", vals, 3)),
                    "status": add_arg(args, "Status", vals, 4),
                }
            )
    else:
        try:
            typ_func = typ_funcs[typ]
            typ_func(sess, action, vals, args)
        except KeyError:
            raise BadRequest(f"The type {typ} is not recognized.")


class GeneralImporter(threading.Thread):
    def __init__(self, f):
        threading.Thread.__init__(self)
        self.line_number = None
        self.f = f
        self.error_message = None
        self.args = []
        self.failed_args = []
        self.failed_lines = []

    def run(self):
        try:
            with Session() as sess:
                reader = csv.reader(self.f)
                hh_data = []
                for idx, line in enumerate(reader):
                    self.args = []
                    self.line_number = idx + 1

                    if len(line) == 0:
                        continue

                    if len(line) > 0 and line[0].startswith("#"):
                        continue

                    try:
                        with sess.begin_nested():
                            _process_line(sess, hh_data, line, self.args)

                    except BadRequest as e:
                        self.args.insert(0, ("Error", e.description))
                        self.failed_args.append(self.args)
                        self.failed_lines.append(line)
                        if len(self.failed_lines) > 1000:
                            break

                HhDatum.insert(sess, hh_data)

                if len(self.failed_lines) == 0:
                    sess.commit()
                else:
                    sess.rollback()
        except BadRequest as e:
            self.error_message = e.description
        except BaseException:
            self.error_message = traceback.format_exc()


def start_process(f):
    with process_lock:
        global process_id
        proc_id = process_id
        process_id += 1

    process = GeneralImporter(f)
    processes[proc_id] = process
    process.start()
    return proc_id


def get_process_ids():
    with process_lock:
        return processes.keys()


def get_process(id):
    return processes[id]



import threading
import traceback

from flask import g, redirect, request

from odio import create_spreadsheet

from sqlalchemy import or_, select, true
from sqlalchemy.orm import joinedload
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

import chellow.e.computer
from chellow.dloads import open_file
from chellow.e.computer import contract_func, forecast_date
from chellow.gas.engine import GDataSource, g_rates
from chellow.models import (
    GBill,
    GContract,
    GEra,
    GSupply,
    RSession,
    ReportRun,
    Scenario,
    Site,
    SiteGEra,
    User,
)
from chellow.utils import (
    PropDict,
    c_months_c,
    c_months_u,
    hh_format,
    hh_max,
    hh_min,
    hh_range,
    make_val,
    req_bool,
    req_int,
    req_str,
    to_utc,
    utc_datetime_now,
)


def write_spreadsheet(fl, compressed, site_rows, era_rows):
    fl.seek(0)
    fl.truncate()
    with create_spreadsheet(fl, compressed=compressed) as sheet:
        sheet.append_table("Site Level", site_rows)
        sheet.append_table("Era Level", era_rows)


def _process_era(
    report_context,
    sess,
    site,
    g_era,
    month_start,
    month_finish,
    forecast_from,
    g_era_rows,
    vb_titles,
    now,
    summary_titles,
):
    g_supply = g_era.g_supply

    ss_start = hh_max(g_era.start_date, month_start)
    ss_finish = hh_min(g_era.finish_date, month_finish)

    ss = GDataSource(
        sess,
        ss_start,
        ss_finish,
        forecast_from,
        g_era,
        report_context,
        None,
    )

    contract = g_era.g_contract
    vb_function = contract_func(report_context, contract, "virtual_bill")
    if vb_function is None:
        raise BadRequest(
            f"The contract {contract.name} doesn't have the virtual_bill() function."
        )
    try:
        vb_function(ss)
        bill = ss.bill
    except TypeError as e:
        raise BadRequest(
            f"Problem with virtual bill for g_supplier_contrat {contract.id}."
        ) from e

    summary_data = {
        "kwh": 0,
        "net_gbp": 0,
        "vat_gbp": 0,
        "gross_gbp": 0,
        "billed_kwh": 0,
        "billed_net_gbp": 0,
        "billed_vat_gbp": 0,
        "billed_gross_gbp": 0,
    }

    for key in ("kwh", "net_gbp", "vat_gbp", "gross_gbp"):
        try:
            summary_data[key] += bill[key]
        except KeyError:
            bill["problem"] += (
                f"For the supply {ss.mprn} the virtual bill {bill} "
                f"from the contract {contract.name} does not contain "
                f"the {key} key."
            )

    associated_site_codes = {
        s.site.code for s in g_era.site_g_eras if not s.is_physical
    }

    for g_bill in sess.query(GBill).filter(
        GBill.g_supply == g_supply,
        GBill.start_date <= ss_finish,
        GBill.finish_date >= ss_start,
    ):
        bill_start = g_bill.start_date
        bill_finish = g_bill.finish_date
        bill_duration = (bill_finish - bill_start).total_seconds() + (30 * 60)
        overlap_duration = (
            min(bill_finish, ss_finish) - max(bill_start, ss_start)
        ).total_seconds() + (30 * 60)
        overlap_proportion = overlap_duration / bill_duration
        summary_data["billed_kwh"] += overlap_proportion * float(g_bill.kwh)
        summary_data["billed_net_gbp"] += overlap_proportion * float(g_bill.net)
        summary_data["billed_vat_gbp"] += overlap_proportion * float(g_bill.vat)
        summary_data["billed_gross_gbp"] += overlap_proportion * float(g_bill.gross)

    g_era_row = (
        [
            make_val(v)
            for v in [
                now,
                g_supply.mprn,
                g_supply.name,
                g_supply.g_exit_zone.code,
                g_era.msn,
                g_era.g_unit.code,
                contract.name,
                site.code,
                site.name,
                associated_site_codes,
                g_era.start_date,
                month_finish,
            ]
        ]
        + [make_val(summary_data[t]) for t in summary_titles]
        + [make_val(bill.get(t)) for t in vb_titles]
    )
    g_era_rows.append(g_era_row)
    return summary_data


def content(scenario_props, user_id, compression, now, base_name):
    report_context = {}
    try:
        eng = report_context["g_engine"]
    except KeyError:
        eng = report_context["g_engine"] = {}

    try:
        rss_cache = eng["rates"]
    except KeyError:
        rss_cache = eng["rates"] = {}

    try:
        with RSession() as sess:
            start_year = scenario_props["scenario_start_year"]
            start_month = scenario_props["scenario_start_month"]
            months = scenario_props["scenario_duration"]

            month_pairs = list(
                c_months_u(
                    start_year=start_year, start_month=start_month, months=months
                )
            )
            start_date = month_pairs[0][0]
            finish_date = month_pairs[-1][-1]

            base_name.append(
                hh_format(start_date)
                .replace(" ", "_")
                .replace(":", "")
                .replace("-", "")
            )

            base_name.append("for")
            base_name.append(str(months))
            base_name.append("months")

            if "forecast_from" in scenario_props:
                forecast_from = scenario_props["forecast_from"]
            else:
                forecast_from = None

            if forecast_from is None:
                forecast_from = forecast_date()
            else:
                forecast_from = to_utc(forecast_from)

            sites = select(Site).join(SiteGEra).distinct().order_by(Site.code)

            mprns = scenario_props.get("mprns")
            g_supply_ids = None
            if mprns is not None:
                g_supply_ids = []
                for mprn in mprns:
                    g_supply = GSupply.get_by_mprn(sess, mprn)
                    g_supply_ids.append(g_supply.id)

                if len(g_supply_ids) == 1:
                    base_name.append("g_supply")
                    base_name.append(str(g_supply.id))
                else:
                    base_name.append("mprns")

                sites = (
                    sites.join(GEra).join(GSupply).where(GSupply.id.in_(g_supply_ids))
                )

            site_codes = scenario_props.get("site_codes")
            if site_codes is not None:
                if len(site_codes) == 1:
                    base_name.append("site")
                    base_name.append(site_codes[0])
                else:
                    base_name.append("sitecodes")
                sites = sites.where(Site.code.in_(site_codes))

            for is_industry, rates_prop in (
                (True, "industry_rates"),
                (False, "supplier_rates"),
            ):

                for rate_script in scenario_props.get(rates_prop, []):
                    g_contract_id = rate_script["g_contract_id"]
                    try:
                        i_cache = rss_cache[is_industry]
                    except KeyError:
                        i_cache = rss_cache[is_industry] = {}

                    try:
                        cont_cache = i_cache[g_contract_id]
                    except KeyError:
                        cont_cache = i_cache[g_contract_id] = {}

                    try:
                        rate_script_start = rate_script["start_date"]
                    except KeyError:
                        raise BadRequest(
                            f"Problem in the scenario properties. Can't find the "
                            f"'start_date' key of the contract {g_contract_id} in the "
                            f"'{rates_prop}' map."
                        )

                    try:
                        rate_script_finish = rate_script["finish_date"]
                    except KeyError:
                        raise BadRequest(
                            f"Problem in the scenario properties. Can't find the "
                            f"'finish_date' key of the contract {g_contract_id} in the "
                            f"'{rates_prop}' map."
                        )

                    for dt in hh_range(
                        report_context, rate_script_start, rate_script_finish
                    ):
                        g_rates(sess, report_context, g_contract_id, dt, is_industry)

                    for dt in hh_range(
                        report_context, rate_script_start, rate_script_finish
                    ):
                        storage = cont_cache[dt]._storage.copy()
                        storage.update(rate_script["script"])
                        cont_cache[dt] = PropDict("scenario properties", storage)

            user = User.get_by_id(sess, user_id)
            fname = "_".join(base_name) + ".ods"
            rf = open_file(fname, user, mode="wb")
            org_rows = []
            site_rows = []
            g_era_rows = []

            org_header_titles = [
                "creation_date",
                "month",
            ]
            summary_titles = [
                "kwh",
                "net_gbp",
                "vat_gbp",
                "gross_gbp",
                "billed_kwh",
                "billed_net_gbp",
                "billed_vat_gbp",
                "billed_gross_gbp",
            ]
            org_titles = org_header_titles + summary_titles
            site_header_titles = [
                "creation_date",
                "site_code",
                "site_name",
                "associated_site_codes",
                "month",
            ]
            site_titles = site_header_titles + summary_titles
            era_header_titles = [
                "creation_date",
                "mprn",
                "supply_name",
                "exit_zone",
                "msn",
                "unit",
                "contract",
                "site_code",
                "site_name",
                "associated_site_codes",
                "era-start",
                "month",
            ]

            vb_titles = []
            conts = (
                sess.query(GContract)
                .join(GEra)
                .join(GSupply)
                .where(
                    GEra.start_date <= finish_date,
                    or_(GEra.finish_date == null(), GEra.finish_date >= start_date),
                )
                .distinct()
                .order_by(GContract.id)
            )
            if g_supply_ids is not None:
                conts = conts.where(GEra.g_supply_id.in_(g_supply_ids))
            for cont in conts:
                title_func = chellow.e.computer.contract_func(
                    report_context, cont, "virtual_bill_titles"
                )
                if title_func is None:
                    raise Exception(
                        f"For the contract {cont.name} there doesn't seem to be "
                        f"a 'virtual_bill_titles' function."
                    )
                for title in title_func():
                    if title not in vb_titles:
                        vb_titles.append(title)

            org_rows.append(org_header_titles + summary_titles)
            site_rows.append(site_header_titles + summary_titles)
            g_era_rows.append(era_header_titles + summary_titles + vb_titles)

            for month_start, month_finish in month_pairs:
                org_row = {
                    "creation_date": now,
                    "month": month_start,
                    "kwh": 0,
                    "net_gbp": 0,
                    "vat_gbp": 0,
                    "gross_gbp": 0,
                    "billed_kwh": 0,
                    "billed_net_gbp": 0,
                    "billed_vat_gbp": 0,
                    "billed_gross_gbp": 0,
                }
                for site in sess.scalars(
                    sites.where(
                        GEra.start_date <= month_finish,
                        or_(
                            GEra.finish_date == null(), GEra.finish_date >= month_start
                        ),
                    )
                ):
                    linked_sites = {
                        s.code
                        for s in site.find_linked_sites(sess, month_start, month_finish)
                    }

                    site_row = {
                        "creation_date": now,
                        "site_code": site.code,
                        "site_name": site.name,
                        "associated_site_codes": linked_sites,
                        "month": month_finish,
                        "kwh": 0,
                        "net_gbp": 0,
                        "vat_gbp": 0,
                        "gross_gbp": 0,
                        "billed_kwh": 0,
                        "billed_net_gbp": 0,
                        "billed_vat_gbp": 0,
                        "billed_gross_gbp": 0,
                    }

                    g_eras_q = (
                        select(GEra)
                        .join(SiteGEra)
                        .filter(
                            SiteGEra.site == site,
                            SiteGEra.is_physical == true(),
                            GEra.start_date <= month_finish,
                            or_(
                                GEra.finish_date == null(),
                                GEra.finish_date >= month_start,
                            ),
                        )
                        .options(
                            joinedload(GEra.g_contract),
                            joinedload(GEra.g_supply),
                            joinedload(GEra.g_supply).joinedload(GSupply.g_exit_zone),
                        )
                        .order_by(GEra.id)
                    )
                    if g_supply_ids is not None:
                        g_eras_q = g_eras_q.where(GEra.g_supply_id.in_(g_supply_ids))

                    for g_era in sess.scalars(g_eras_q):
                        try:
                            summary_data = _process_era(
                                report_context,
                                sess,
                                site,
                                g_era,
                                month_start,
                                month_finish,
                                forecast_from,
                                g_era_rows,
                                vb_titles,
                                now,
                                summary_titles,
                            )
                        except BadRequest as e:
                            raise BadRequest(
                                f"Problem with g_era {g_era.id}: {e.description}"
                            )
                        for k, v in summary_data.items():
                            org_row[k] += v
                            site_row[k] += v

                    site_rows.append([make_val(site_row[t]) for t in site_titles])
                    sess.rollback()
                org_rows.append([make_val(org_row[t]) for t in org_titles])
                write_spreadsheet(rf, compression, site_rows, g_era_rows)

        if scenario_props.get("save_report_run", False):
            report_run_id = ReportRun.w_insert(
                "g_monthly_duration", user_id, fname, {"scenario": scenario_props}
            )
            for tab, rows in (
                ("org", org_rows),
                ("site", site_rows),
                ("era", g_era_rows),
            ):
                titles = rows[0]
                for row in rows[1:]:
                    values = dict(zip(titles, row))
                    ReportRun.w_insert_row(report_run_id, tab, titles, values, {})
            ReportRun.w_update(report_run_id, "finished")

    except BaseException:
        msg = traceback.format_exc()
        print(msg)
        site_rows.append(["Problem " + msg])
        write_spreadsheet(rf, compression, site_rows, g_era_rows)
    finally:
        try:
            rf.close()
        except BaseException:
            msg = traceback.format_exc()
            ef = open_file("error.txt", user, mode="w")
            ef.write(msg + "\n")
            ef.close()


def do_get(sess):
    now = utc_datetime_now()
    base_name = []
    if "scenario_id" in request.values:
        scenario_id = req_int("scenario_id")
        scenario = Scenario.get_by_id(sess, scenario_id)
        scenario_props = scenario.props
        base_name.append(scenario.name)
    else:
        scenario_props = {}
        base_name.append("g_monthly_duration")

    if "finish_year" in request.values:
        year = req_int("finish_year")
        month = req_int("finish_month")
        months = req_int("months")
        start_date, _ = next(
            c_months_c(finish_year=year, finish_month=month, months=months)
        )
        scenario_props["scenario_start_year"] = start_date.year
        scenario_props["scenario_start_month"] = start_date.month
        scenario_props["scenario_duration"] = months

    if "site_id" in request.values:
        site_id = req_int("site_id")
        scenario_props["site_codes"] = [Site.get_by_id(sess, site_id).code]

    if "site_codes" in request.values:
        site_codes_raw_str = req_str("site_codes")
        site_codes_str = site_codes_raw_str.strip()
        if len(site_codes_str) > 0:
            site_codes = []

            for site_code in site_codes_str.splitlines():
                Site.get_by_code(sess, site_code)  # Check valid
                site_codes.append(site_code)

            scenario_props["site_codes"] = site_codes

    if "g_supply_id" in request.values:
        g_supply_id = req_int("g_supply_id")
        g_supply = GSupply.get_by_id(sess, g_supply_id)
        scenario_props["mprns"] = [g_supply.mprn]

    if "compression" in request.values:
        compression = req_bool("compression")
    else:
        compression = True

    args = scenario_props, g.user.id, compression, now, base_name

    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



import csv
from datetime import datetime as Datetime
from decimal import Decimal
from io import StringIO

from dateutil.relativedelta import relativedelta

from werkzeug.exceptions import BadRequest

from chellow.utils import HH, parse_mpan_core, to_utc, validate_hh_start

col_map = {
    "Gsp Data": "sum-gsp-kwh",
    "Levy Exempt Energy Units": "lec-kwh",
    "Levy Exempt Energy Price": "lec-rate",
    "Levy Exempt Energy Charge": "lec-gbp",
    "Meter Reading Charges Charge": "meter-reading-gbp",
    "Availability Charges Units": "duos-availability-kva",
    "Availability Charges Price": "duos-availability-rate",
    "Availability Charges Charge": "duos-availability-gbp",
    "Excess Availability Charges Units": "duos-excess-availability-kva",
    "Excess Availability Charges Price": "duos-excess-availability-rate",
    "Excess Availability Charges Charge": "duos-excess-availability-gbp",
    "Fixed Charges Units": "duos-fixed-days",
    "Fixed Charges Price": "duos-fixed-rate",
    "Fixed Charges Charge": "duos-fixed-gbp",
    "Rate 1 Units": "duos-red-kwh",
    "Rate 1 Price": "duos-red-rate",
    "Rate 1 Charge": "duos-red-gbp",
    "Rate 2 Units": "duos-amber-kwh",
    "Rate 2 Price": "duos-amber-rate",
    "Rate 2 Charge": "duos-amber-gbp",
    "Rate 3 Units": "duos-green-kwh",
    "Rate 3 Price": "duos-green-rate",
    "Rate 3 Charge": "duos-green-gbp",
    "Reactive Power Charge Units": "duos-reactive-kvarh",
    "Reactive Power Charge Price": "duos-reactive-rate",
    "Reactive Power Charge Charge": "duos-reactive-gbp",
    "BSUoS Actual Charge (Date) Charge": "bsuos-actual-gbp-info",
    "BSUOSTTL": "bsuos-gbp",
    "AAHEDC Charge Units": "aahedc-gsp-kwh",
    "AAHEDC Charge Price": "aahedc-rate",
    "AAHEDC Charge Charge": "aahedc-gbp",
    "RO Adjustment Units": "ro-msp-kwh",
    "RO Adjustment Price": "ro-rate",
    "RO Adjustment Charge": "ro-gbp",
    "Estimated FiT Charge Units": "fit-estimate-msp-kwh",
    "Estimated FiT Charge Price": "fit-estimate-rate",
    "Estimated FiT Charge Charge": "fit-estimate-gbp",
    "Actual FiT Charge Price": "fit-previous-actual-rate",
    "Actual FiT Charge Charge": "fit-previous-actual-gbp",
    "Estimated CfD Charge Units": "cfd-fit-estimate-nbp-kwh",
    "Estimated CfD Charge Price": "cfd-fit-estimate-rate",
    "Estimated CfD Charge Charge": "cfd-fit-estimate-gbp",
    "Actual CfD Charge Units": "cfd-fit-previous-actual-nbp-kwh",
    "Actual CfD Charge Price": "cfd-fit-previous-actual-rate",
    "Actual CfD Charge Charge": "cfd-fit-previous-actual-gbp",
    "CFD_Total": "cfd-fit-total",
    "AHC_NRG1": "reconciliation-gbp",
    "Energy Peak Units": "peak-gsp-kwh",
    "Energy Peak Price": "peak-rate",
    "Energy Peak Charge": "peak-gbp",
    "Energy Peak Shoulder Units": "peak-shoulder-gsp-kwh",
    "Energy Peak Shoulder Price": "peak-shoulder-rate",
    "Energy Peak Shoulder Charge": "peak-shoulder-gbp",
    "Energy Summer Night Units": "summer-night-gsp-kwh",
    "Energy Summer Night Price": "summer-night-rate",
    "Energy Summer Night Charge": "summer-night-gbp",
    "Energy Summer Weekday Units": "summer-weekday-gsp-kwh",
    "Energy Summer Weekday Price": "summer-weekday-rate",
    "Energy Summer Weekday Charge": "summer-weekday-gbp",
    "Energy Summer Weekend Units": "summer-weekend-gsp-kwh",
    "Energy Summer Weekend Price": "summer-weekend-rate",
    "Energy Summer Weekend Charge": "summer-weekend-gbp",
    "Energy Winter Night Units": "winter-night-gsp-kwh",
    "Energy Winter Night Price": "winter-night-rate",
    "Energy Winter Night Charge": "winter-night-gbp",
    "Energy Winter Weekday Units": "winter-weekday-gsp-kwh",
    "Energy Winter Weekday Price": "winter-weekday-rate",
    "Energy Winter Weekday Charge": "winter-weekday-gbp",
    "Energy Winter Weekend Units": "winter-weekend-gsp-kwh",
    "Energy Winter Weekend Price": "winter-weekend-rate",
    "Energy Winter Weekend Charge": "winter-weekend-gbp",
    "Network UoS Charges Price": "triad-estimate-rate",
    "Network UoS Charges Units": "triad-estimate-gsp-kw",
    "Network UoS Charges Charge": "triad-estimate-gbp",
    "CAP_M_TTL": "capacity-market-gbp",
}


class Parser:
    def __init__(self, f):
        self.last_line = None
        tf = StringIO(str(f.read(), "utf-8", errors="ignore"))
        lines = (self._set_last_line(i, l) for i, l in enumerate(tf))
        self.reader = csv.reader(lines, skipinitialspace=True)
        self._line_number = None
        self._title_line = None

    @property
    def line_number(self):
        return None if self._line_number is None else self._line_number + 1

    def _set_last_line(self, i, line):
        self._line_numer = i
        self.last_line = line
        if i == 0:
            self._title_line = line
        return line

    def make_raw_bills(self):
        titles = [t.strip() for t in next(iter(self.reader))]
        title_idx = dict((title, i) for i, title in enumerate(titles))

        raw_bills = []

        for vals in self.reader:
            if len(vals) == 0 or vals[0].startswith("#") or "".join(vals) == "":
                continue

            def val(title):
                idx = title_idx[title]
                try:
                    return vals[idx]
                except KeyError:
                    raise BadRequest("For title " + title + " and index " + str(idx))

            def date_val(title):
                try:
                    return to_utc(Datetime.strptime(val(title), "%d/%m/%Y"))
                except ValueError as e:
                    raise BadRequest(
                        "At line number "
                        + str(self._line_number)
                        + ", while trying to find the value of "
                        + title
                        + " the date couldn't be parsed. "
                        + str(e)
                        + " The full line is "
                        + self.last_line
                    )

            def dec_val(title):
                return Decimal(val(title))

            issue_date = date_val("Invoice Date")
            bill_from = validate_hh_start(date_val("Bill Period Start"))
            bill_to = (
                validate_hh_start(date_val("Bill Period End"))
                + relativedelta(days=1)
                - HH
            )
            kwh = dec_val("Site Data")
            breakdown = dict(
                [(v, dec_val(k)) for k, v in col_map.items() if k in titles]
            )
            breakdown["raw_lines"] = [self._title_line.strip(), self.last_line.strip()]
            net = Decimal("0.00") + dec_val("NET_AMT")
            vat = Decimal("0.00") + dec_val("VATTTL")
            gross = Decimal("0.00") + dec_val("TTLCHG")
            mpan_core = parse_mpan_core(val("Mpan"))
            raw_bill = {
                "bill_type_code": "N",
                "account": val("Bill Ref No."),
                "mpan_core": mpan_core,
                "reference": val("Invoice No."),
                "issue_date": issue_date,
                "start_date": bill_from,
                "finish_date": bill_to,
                "kwh": kwh,
                "net": net,
                "vat": vat,
                "gross": gross,
                "breakdown": breakdown,
                "reads": [],
            }
            raw_bills.append(raw_bill)

        return raw_bills



import csv
import threading
import traceback
from datetime import datetime as Datetime

from dateutil.relativedelta import relativedelta

from flask import g, redirect

import pytz

from sqlalchemy.sql.expression import null, or_, true

from werkzeug.exceptions import BadRequest

from chellow.dloads import open_file
from chellow.e.computer import SupplySource, contract_func, forecast_date
from chellow.models import Era, Session, Site, SiteEra, Supply, User
from chellow.utils import HH, csv_make_val, hh_format, hh_max, hh_min, req_bool, req_int


def content(
    start_year,
    start_month,
    start_day,
    finish_year,
    finish_month,
    finish_day,
    is_import,
    supply_id,
    user_id,
):
    caches = {}
    try:
        with Session() as sess:
            user = User.get_by_id(sess, user_id)
            f = open_file("daily_supplier_virtual_bill.csv", user, mode="w", newline="")
            writer = csv.writer(f, lineterminator="\n")
            start_date = Datetime(start_year, start_month, start_day, tzinfo=pytz.utc)
            finish_date = (
                Datetime(finish_year, finish_month, finish_day, tzinfo=pytz.utc)
                + relativedelta(days=1)
                - HH
            )

            supply = Supply.get_by_id(sess, supply_id)
            fd = forecast_date()
            day_start = start_date
            header_titles = [
                "MPAN Core",
                "Site Code",
                "Site Name",
                "Account",
                "From",
                "To",
                "Is Forecast?",
            ]

            bill_titles = []
            # Find titles
            for era in sess.query(Era).filter(
                Era.supply == supply,
                Era.start_date <= finish_date,
                or_(Era.finish_date == null(), Era.finish_date >= start_date),
            ):
                if is_import:
                    cont = era.imp_supplier_contract
                else:
                    cont = era.exp_supplier_contract

                for title in contract_func(caches, cont, "virtual_bill_titles")():
                    if title not in bill_titles:
                        bill_titles.append(title)

                ssc = era.ssc
                if ssc is not None:
                    for mr in ssc.measurement_requirements:
                        for suffix in ("-kwh", "-rate", "-gbp"):
                            title = mr.tpr.code + suffix
                            if title not in bill_titles:
                                bill_titles.append(title)

            writer.writerow(header_titles + bill_titles)

            while not day_start > finish_date:
                day_finish = day_start + relativedelta(days=1) - HH

                for era in supply.find_eras(sess, day_start, day_finish):
                    chunk_start = hh_max(era.start_date, day_start)
                    chunk_finish = hh_min(era.finish_date, day_finish)

                    ss = SupplySource(
                        sess,
                        chunk_start,
                        chunk_finish,
                        fd,
                        era,
                        is_import,
                        caches,
                    )

                    site = (
                        sess.query(Site)
                        .join(SiteEra)
                        .filter(SiteEra.era == era, SiteEra.is_physical == true())
                        .one()
                    )
                    row = [
                        ss.mpan_core,
                        site.code,
                        site.name,
                        ss.supplier_account,
                        hh_format(ss.start_date),
                        hh_format(ss.finish_date),
                        ss.years_back > 0,
                    ]

                    contract_func(caches, ss.supplier_contract, "virtual_bill")(ss)
                    bill = ss.supplier_bill
                    for title in bill_titles:
                        if title in bill:
                            row.append(csv_make_val(bill[title]))
                            del bill[title]
                        else:
                            row.append("")

                    for k in sorted(bill.keys()):
                        row.append(k)
                        row.append(csv_make_val(bill[k]))
                    writer.writerow(row)

                day_start += relativedelta(days=1)
    except BadRequest as e:
        writer.writerow(["Problem: " + e.description])
    except BaseException:
        writer.writerow([traceback.format_exc()])
    finally:
        if f is not None:
            f.close()


def do_get(sess):
    start_year = req_int("start_year")
    start_month = req_int("start_month")
    start_day = req_int("start_day")

    finish_year = req_int("finish_year")
    finish_month = req_int("finish_month")
    finish_day = req_int("finish_day")

    is_import = req_bool("is_import")
    supply_id = req_int("supply_id")

    args = (
        start_year,
        start_month,
        start_day,
        finish_year,
        finish_month,
        finish_day,
        is_import,
        supply_id,
        g.user.id,
    )
    threading.Thread(target=content, args=args).start()
    return redirect("/downloads", 303)



from utils import match

from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_sources,
    insert_voltage_levels,
)
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_supplies_get(sess, client):
    sess.commit()

    query_string = {
        "start_date": "2024-05-01T00:00:00Z",
        "finish_date": "2024-05-01T00:00:00Z",
    }
    response = client.get("/api/v1/supplies", query_string=query_string)
    match(response, 200)


def test_channel_get(sess, client):
    vf = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.get_by_code(sess, "Z")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "None core", vf, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess, "bank_holidays", "", {}, vf, None, bank_holiday_rate_script
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", vf, None, None)
    participant.insert_party(sess, market_role_X, "Fusion Ltc", vf, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", vf, None, None)
    mop_contract = Contract.insert_mop(
        sess, "Fusion Mop Contract", participant, "", {}, vf, None, {}
    )
    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, "", {}, vf, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", vf, None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    imp_supplier_contract = Contract.insert_supplier(
        sess, "Fusion Supplier 2000", participant, "", {}, vf, None, {}
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", vf, None, "22")
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", vf, None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", vf, None)
    mtc = Mtc.insert(sess, "845", False, True, vf, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        vf,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess, "510", "PC 5-8 & HH HV", voltage_level, False, True, vf, None
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, vf, None)
    insert_sources(sess)
    source = Source.get_by_code(sess, "grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    supply = site.insert_e_supply(
        sess,
        source,
        None,
        "Bob",
        utc_datetime(2000, 1, 1),
        utc_datetime(2000, 1, 31, 23, 30),
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )
    era = supply.eras[0]
    channel = era.insert_channel(sess, True, "ACTIVE")
    sess.commit()

    query_string = {
        "start_date": "2024-05-01T00:00:00Z",
        "finish_date": "2024-05-01T00:00:00Z",
    }
    response = client.get(f"/api/v1/channel/{channel.id}", query_string=query_string)
    match(response, 200)



from chellow.e.system_price import elexon_import
from chellow.models import Contract, MarketRole, Participant
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_process(sess, mocker):
    from_date = to_utc(ct_datetime(2000, 1, 1))
    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AK Industries")
    participant.insert_party(sess, market_role_Z, "non-core", from_date, None, None)
    system_price_properties = {
        "enabled": True,
    }
    rate_script_properties = {"gbp_per_nbp_mwh": {}}
    Contract.insert_non_core(
        sess,
        "system_price",
        "",
        system_price_properties,
        utc_datetime(2001, 4, 1),
        None,
        rate_script_properties,
    )
    Contract.insert_non_core(
        sess,
        "configuration",
        "",
        {"elexonportal_scripting_key": "xxx"},
        from_date,
        None,
        {},
    )
    sess.commit()

    log = []

    def log_f(msg):
        log.append(msg)

    mock_response = mocker.Mock()
    with open("test/e/system_price/prices.xls", "rb") as f:
        mock_response.content = f.read()
    mock_response.status_code = 200
    mock_response.reason = "OK"
    mocker.patch("chellow.e.system_price.download_file", return_value=mock_response)

    mock_set_progress = mocker.Mock()
    scripting_key = "xxx"
    elexon_import(sess, log_f, mock_set_progress, scripting_key)

    assert log == [
        "Starting to check System Prices.",
        "Extracting data from 2001-04-01 01:00",
        "Received 200 OK",
        "Successfully extracted data.",
        "Updating rate script starting at 2001-04-01 01:00.",
    ]



import chellow.dloads
from chellow.models import User, UserRole
from chellow.reports.report_sscs import content


def test_content(sess):
    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)
    user_id = user.id
    sess.commit()

    compression = False
    content(user_id, compression)

    files = list(p.name for p in chellow.dloads.download_path.iterdir())
    assert files == ["00000_FINISHED_admin_example_com_sscs.ods"]



import csv
from datetime import datetime as Datetime
from decimal import Decimal, InvalidOperation
from io import StringIO

from dateutil.relativedelta import relativedelta


from sqlalchemy import or_, select
from sqlalchemy.sql.expression import null

from werkzeug.exceptions import BadRequest

from zish import dumps, loads

from chellow.e.elexon import download_file
from chellow.models import Contract, RateScript
from chellow.rate_server import download
from chellow.utils import HH, ct_datetime, hh_format, hh_range, to_ct, to_utc

RUNS = ["DF", "RF", "R3", "R2", "R1", "SF", "II"]


def key_format(dt):
    return dt.strftime("%d %H:%M")


def hh(data_source, run="DF"):
    gsp_group_code = data_source.gsp_group_code
    use_runs = RUNS[RUNS.index(run) :]

    try:
        cache = data_source.caches["tlms"]
    except KeyError:
        cache = data_source.caches["tlms"] = {}

    for h in data_source.hh_data:
        try:
            h["tlm"] = tlm = cache[h["start-date"]][gsp_group_code][run]
        except KeyError:
            h_start = h["start-date"]
            vals = data_source.non_core_rate("tlms", h_start, exact=True)
            if vals is None:
                vals = data_source.non_core_rate("tlms", h["hist-start"])

            rates = vals["tlms"]

            key = key_format(h_start)
            try:
                rate = rates[key]
            except KeyError:
                rate = sorted(rates.items())[-1][1]

            try:
                gsp_rate = rate[gsp_group_code]
            except KeyError:
                gsp_rate = sorted(rate.items())[-1][1]

            tlm = None
            for use_run in use_runs:
                try:
                    h["tlm"] = tlm = float(gsp_rate[use_run]["off_taking"])
                    break
                except KeyError:
                    pass

            if tlm is None:
                h["tlm"] = tlm = float(sorted(gsp_rate.items())[-1][1]["off_taking"])

            try:
                rates_cache = cache[h["start-date"]]
            except KeyError:
                rates_cache = cache[h["start-date"]] = {}

            try:
                gsp_cache = rates_cache[gsp_group_code]
            except KeyError:
                gsp_cache = rates_cache[gsp_group_code] = {}

            gsp_cache[run] = tlm

        h["nbp-kwh"] = h["gsp-kwh"] * tlm


def _find_complete_date(caches, sess, contract, cache):
    for rs in sess.scalars(
        select(RateScript)
        .where(RateScript.contract == contract, RateScript.finish_date != null())
        .order_by(RateScript.start_date.desc())
    ):
        rates = rs.make_script()
        rates["id"] = rs.id
        cache["rate_scripts"].append(rates)
        timestamps = cache["timestamps"]
        try:
            tlms = rates["tlms"]
        except KeyError:
            tlms = rates["tlms"] = {}
        complete = True
        for dt in hh_range(caches, rs.start_date, rs.finish_date):
            timestamps[dt] = rates
            key = key_format(dt)
            if key in tlms:
                gps = tlms[key]
                for group in GSP_GROUP_LOOKUP.values():
                    if not (group in gps and "DF" in gps[group]):
                        complete = False
            else:
                complete = False

        if complete:
            return rs.finish_date


def elexon_import(sess, log, set_progress, scripting_key):
    import_tlms(sess, log, set_progress, scripting_key)
    # import_etlms(sess, log, set_progress, s)


def import_tlms(sess, log, set_progress, scripting_key):
    cache = {"rate_scripts": [], "timestamps": {}}
    caches = {}
    log("Starting to check TLMs.")
    contract_name = "tlms"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        contract = Contract.insert_non_core(
            sess,
            contract_name,
            "",
            {"enabled": True},
            to_utc(ct_datetime(1997, 1, 1)),
            to_utc(ct_datetime(1997, 1, 31, 23, 30)),
            {"tlms": {}},
        )
        sess.commit()
    contract_props = contract.make_properties()
    if contract_props.get("enabled", True):
        complete_date = _find_complete_date(caches, sess, contract, cache)
        log(f"Found complete up to {complete_date}")

        sess.rollback()  # Avoid long-running transaction
        r = download_file(log, scripting_key, "TLM_FILE")
        parser = csv.reader(
            (x.decode() for x in r.iter_lines()), delimiter=",", quotechar='"'
        )

        next(parser, None)
        for values in parser:
            if len(values) == 0:
                continue

            set_progress(values[0])

            if len(values) > 5:
                _process_line(cache, sess, contract, log, values, complete_date, caches)

        for rscript in cache["rate_scripts"]:
            rs = sess.scalars(
                select(RateScript).where(RateScript.id == rscript["id"])
            ).one()
            rs.script = dumps({"tlms": rscript["tlms"]})
            sess.commit()

    else:
        log(
            "The importer is disabled. Set 'enabled' to 'true' in the "
            "properties to enable it."
        )
    log("Finished checking TLMs.")


GSP_GROUP_LOOKUP = {
    "1": "_A",
    "2": "_B",
    "3": "_C",
    "4": "_D",
    "5": "_E",
    "6": "_F",
    "7": "_G",
    "8": "_H",
    "9": "_J",
    "10": "_K",
    "11": "_L",
    "12": "_M",
    "13": "_N",
    "14": "_P",
}


def _process_line(cache, sess, contract, log_func, values, complete_date, caches):
    hh_date_ct = to_ct(Datetime.strptime(values[0], "%d/%m/%Y"))
    hh_date = to_utc(hh_date_ct)
    hh_date += relativedelta(minutes=30 * (int(values[2]) - 1))
    if complete_date is not None and hh_date <= complete_date:
        return
    run = values[1]
    gsp_group_code = GSP_GROUP_LOOKUP[values[3]]
    off_taking_str = values[4]

    try:
        off_taking = Decimal(off_taking_str)
    except InvalidOperation as e:
        raise BadRequest(
            f"Problem parsing 'off-taking' field '{off_taking_str}' in the row "
            f"{values}. {e}"
        )

    delivering = Decimal(values[5])

    key = key_format(hh_date)
    try:
        cache["timestamps"][hh_date]["tlms"][key][gsp_group_code][run]
    except KeyError:
        try:
            rate = cache["timestamps"][hh_date]
        except KeyError:

            rs = (
                sess.query(RateScript)
                .filter(
                    RateScript.contract == contract,
                    RateScript.start_date <= hh_date,
                    or_(
                        RateScript.finish_date == null(),
                        RateScript.finish_date >= hh_date,
                    ),
                )
                .first()
            )
            assert rs is None
            while rs is None:
                log_func(f"There's no rate script at {hh_format(hh_date)}.")
                latest_rs = (
                    sess.query(RateScript)
                    .filter(RateScript.contract == contract)
                    .order_by(RateScript.start_date.desc())
                    .first()
                )
                if hh_date < latest_rs.start_date:
                    raise BadRequest(
                        f"The start of the latest rate script must be before "
                        f"{hh_format(hh_date)}"
                    )
                contract.update_rate_script(
                    sess,
                    latest_rs,
                    latest_rs.start_date,
                    latest_rs.start_date + relativedelta(months=2) - HH,
                    loads(latest_rs.script),
                )
                new_rs_start = latest_rs.start_date + relativedelta(months=1)
                new_rs = contract.insert_rate_script(sess, new_rs_start, {})
                rt = {"tlms": {}, "id": new_rs.id}
                cache["rate_scripts"].append(rt)
                for h in hh_range(caches, new_rs.start_date, new_rs.finish_date):
                    cache["timestamps"][h] = rt
                sess.commit()
                log_func(f"Added a rate script starting at {hh_format(new_rs_start)}.")

                rs = (
                    sess.query(RateScript)
                    .filter(
                        RateScript.contract == contract,
                        RateScript.start_date <= hh_date,
                        or_(
                            RateScript.finish_date == null(),
                            RateScript.finish_date >= hh_date,
                        ),
                    )
                    .first()
                )

            rate = cache["timestamps"][hh_date]

        try:
            existing = rate["tlms"][key]
        except KeyError:
            existing = rate["tlms"][key] = {}

        try:
            group = existing[gsp_group_code]
        except KeyError:
            group = existing[gsp_group_code] = {}

        group[run] = {"off_taking": off_taking, "delivering": delivering}

        log_func(
            f"Found rate at {hh_format(hh_date)} for GSP Group {gsp_group_code} "
            f"and run {run}."
        )


"""
def import_etlms(sess, log, set_progress, s):
    log(
        "Starting to check for new ETLMs https://www.elexon.co.uk/bsc/operational/"
        "transmission-losses/estimated-transmission-loss-adjustments/"
    )

    contract_name = "etlms"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        log(f"Can't find contract {contract_name} so creating it.")
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )
        log(f"Created contract {contract.name}.")
    url_str = (
        "https://www.elexon.co.uk/bsc/documents/operations-settlement/"
        "transmission-losses/estimated-transmission-loss-adjustments-data/"
    )
    r = s.get(url_str, timeout=120)
    if r.status_code != 200:
        log(r.text)
        raise BadRequest(
            f"Problem retrieving URL. Expexted status code 200, but received a "
            f"{r.status_code}"
        )

    parser = csv.reader(
        (x.decode() for x in r.iter_lines()), delimiter=",", quotechar='"'
    )
    log(f"Opened {url_str}")

    next(parser, None)
    for values in parser:
        log(f"looking at {values}")
        if len(values) < 4:
            continue

        start_date_str = values[0]
        set_progress(start_date_str)
        start_date = to_utc(to_ct(Datetime.strptime(start_date_str, "%d/%m/%Y")))

        delivering = Decimal(values[2])
        off_taking = Decimal(values[3])

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == start_date,
            )
        ).scalar_one_or_none()
        log(f"Found RS {rs}")
        if rs is None:
            log("RS is none")
            rs = contract.insert_rate_script(sess, start_date, {})

            rs_script = rs.make_script()
            rs_script["delivering"] = delivering
            rs_script["off_taking"] = off_taking
            rs.update(rs_script)
            sess.commit()
            log("Commited")
    log("Finished ETLMs")
"""


def rate_server_import(sess, log, set_progress, paths, with_88_99=True):
    CONFIG_KEY = "etlm_version"
    log("Starting to check for a new ETLM version")
    contract_name = "etlms"
    contract = Contract.find_non_core_by_name(sess, contract_name)
    if contract is None:
        log(f"Can't find contract {contract_name} so creating it.")
        contract = Contract.insert_non_core(
            sess, contract_name, "", {}, to_utc(ct_datetime(1996, 4, 1)), None, {}
        )
        log(f"Created contract {contract.name}.")
    etlm_entries = {}
    for path, url in paths:
        if len(path) == 4:
            year_str, utility, rate_type, file_name = path

            if utility == "electricity" and rate_type == "etlms":
                etlm_entries[f"{year_str}_{file_name}"] = url

    if len(etlm_entries) == 0:
        raise BadRequest("Can't find any ETLM files on the rate server.")

    version_key, url = sorted(etlm_entries.items())[-1]
    log(f"Latest version on rate server: {version_key}.")

    config = Contract.get_non_core_by_name(sess, "configuration")
    state = config.make_state()
    current_version = state.get(CONFIG_KEY)

    log(f"Latest version in Chellow: {current_version}")
    if version_key == current_version:
        return

    log("Starting to import ETLMs")

    csv_file = StringIO(download(url).decode("utf8"))
    csv_reader = iter(csv.reader(csv_file))
    next(csv_reader)  # Skip titles
    for values in csv_reader:
        if len(values) < 4:
            continue

        start_date_str = values[0]
        set_progress(start_date_str)
        start_date = to_utc(to_ct(Datetime.strptime(start_date_str, "%d/%m/%Y")))

        delivering = Decimal(values[2])
        off_taking = Decimal(values[3])

        rs = sess.execute(
            select(RateScript).where(
                RateScript.contract == contract,
                RateScript.start_date == start_date,
            )
        ).scalar_one_or_none()
        if rs is None:
            rs = contract.insert_rate_script(sess, start_date, {})

            rs_script = rs.make_script()
            rs_script["delivering"] = delivering
            rs_script["off_taking"] = off_taking
            rs.update(rs_script)
            sess.commit()

    log("Finished ETLMs")
    config = Contract.get_non_core_by_name(sess, "configuration")
    state = config.make_state()
    state[CONFIG_KEY] = version_key
    config.update_state(state)
    sess.commit()



from io import StringIO

from chellow.models import (
    Comm,
    Contract,
    Cop,
    DtcMeterType,
    EnergisationStatus,
    GeneratorType,
    GspGroup,
    MarketRole,
    MeterPaymentType,
    MeterType,
    Mtc,
    MtcLlfc,
    MtcParticipant,
    Participant,
    Pc,
    Site,
    Source,
    User,
    UserRole,
    VoltageLevel,
    insert_comms,
    insert_cops,
    insert_dtc_meter_types,
    insert_energisation_statuses,
    insert_generator_types,
    insert_sources,
    insert_voltage_levels,
)
from chellow.reports.report_109 import content
from chellow.utils import ct_datetime, to_utc, utc_datetime


def test_content(mocker, sess):
    valid_from = to_utc(ct_datetime(1996, 1, 1))
    site = Site.insert(sess, "CI017", "Water Works")

    market_role_Z = MarketRole.insert(sess, "Z", "Non-core")
    participant = Participant.insert(sess, "CALB", "AB Industries")
    participant.insert_party(sess, market_role_Z, "None core", valid_from, None, None)
    bank_holiday_rate_script = {"bank_holidays": []}
    Contract.insert_non_core(
        sess,
        "bank_holidays",
        "",
        {},
        valid_from,
        None,
        bank_holiday_rate_script,
    )
    market_role_X = MarketRole.insert(sess, "X", "Supplier")
    market_role_M = MarketRole.insert(sess, "M", "Mop")
    market_role_C = MarketRole.insert(sess, "C", "HH Dc")
    market_role_R = MarketRole.insert(sess, "R", "Distributor")
    participant.insert_party(sess, market_role_M, "Fusion Mop", valid_from, None, None)
    participant.insert_party(sess, market_role_X, "Fusion", valid_from, None, None)
    participant.insert_party(sess, market_role_C, "Fusion DC", valid_from, None, None)

    mop_charge_script = """
from chellow.utils import reduce_bill_hhs

def virtual_bill_titles():
    return ['net-gbp', 'problem']

def virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]

        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))

    ds.mop_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    mop_contract = Contract.insert_mop(
        sess,
        "Fusion Mop Contract",
        participant,
        mop_charge_script,
        {},
        valid_from,
        None,
        {},
    )

    dc_charge_script = """
from chellow.utils import reduce_bill_hhs

def virtual_bill_titles():
    return ['net-gbp', 'problem']

def virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]

        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))

    ds.dc_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""

    dc_contract = Contract.insert_dc(
        sess, "Fusion DC 2000", participant, dc_charge_script, {}, valid_from, None, {}
    )
    pc = Pc.insert(sess, "00", "hh", utc_datetime(2000, 1, 1), None)
    insert_cops(sess)
    cop = Cop.get_by_code(sess, "5")
    insert_comms(sess)
    comm = Comm.get_by_code(sess, "GSM")

    supplier_charge_script = """
import chellow.e.ccl
from chellow.utils import HH, reduce_bill_hhs, utc_datetime

def displaced_virtual_bill_titles():
    return [
        'ccl-kwh', 'ccl-rate', 'ccl-gbp', 'net-gbp', 'vat-gbp', 'gross-gbp',
        'sum-msp-kwh', 'sum-msp-gbp', 'problem']

def displaced_virtual_bill(ds):
    for hh in ds.hh_data:
        hh_start = hh['start-date']
        bill_hh = ds.supplier_bill_hhs[hh_start]
        bill_hh['sum-msp-kwh'] = hh['msp-kwh']
        bill_hh['sum-msp-gbp'] = hh['msp-kwh'] * 0.1
        bill_hh['net-gbp'] = sum(
            v for k, v in bill_hh.items() if k.endswith('gbp'))
        bill_hh['vat-gbp'] = 0
        bill_hh['gross-gbp'] = bill_hh['net-gbp'] + bill_hh['vat-gbp']

    ds.supplier_bill = reduce_bill_hhs(ds.supplier_bill_hhs)
"""
    imp_supplier_contract = Contract.insert_supplier(
        sess,
        "Fusion Supplier 2000",
        participant,
        supplier_charge_script,
        {},
        valid_from,
        None,
        {},
    )
    dno = participant.insert_party(sess, market_role_R, "WPD", valid_from, None, "22")
    Contract.insert_dno(sess, dno.dno_code, participant, "", {}, valid_from, None, {})
    meter_type = MeterType.insert(sess, "C5", "COP 1-5", utc_datetime(2000, 1, 1), None)
    meter_payment_type = MeterPaymentType.insert(sess, "CR", "Credit", valid_from, None)
    mtc = Mtc.insert(sess, "845", False, True, valid_from, None)
    mtc_participant = MtcParticipant.insert(
        sess,
        mtc,
        participant,
        "HH COP5 And Above With Comms",
        False,
        True,
        meter_type,
        meter_payment_type,
        0,
        valid_from,
        None,
    )
    insert_voltage_levels(sess)
    voltage_level = VoltageLevel.get_by_code(sess, "HV")
    llfc = dno.insert_llfc(
        sess,
        "510",
        "PC 5-8 & HH HV",
        voltage_level,
        False,
        True,
        valid_from,
        None,
    )
    MtcLlfc.insert(sess, mtc_participant, llfc, valid_from, None)
    insert_sources(sess)
    insert_generator_types(sess)
    generator_type = GeneratorType.get_by_code(sess, "turb")
    source = Source.get_by_code(sess, "gen-grid")
    gsp_group = GspGroup.insert(sess, "_L", "South Western")
    insert_energisation_statuses(sess)
    energisation_status = EnergisationStatus.get_by_code(sess, "E")
    insert_dtc_meter_types(sess)
    dtc_meter_type = DtcMeterType.get_by_code(sess, "H")
    site.insert_e_supply(
        sess,
        source,
        generator_type,
        "Bob",
        valid_from,
        None,
        gsp_group,
        mop_contract,
        dc_contract,
        "hgjeyhuw",
        dno,
        pc,
        "845",
        cop,
        comm,
        None,
        energisation_status,
        dtc_meter_type,
        "22 7867 6232 781",
        "510",
        imp_supplier_contract,
        "7748",
        361,
        None,
        None,
        None,
        None,
        None,
    )

    editor = UserRole.insert(sess, "editor")
    user = User.insert(sess, "admin@example.com", "xxx", editor, None)

    sess.commit()

    mock_file = StringIO()
    mock_file.close = mocker.Mock()
    mocker.patch("chellow.reports.report_109.open_file", return_value=mock_file)

    end_year = 2000
    end_month = 1
    months = 1

    content(imp_supplier_contract.id, end_year, end_month, months, user)

    actual = mock_file.getvalue()
    print(actual)

    expected = [
        [
            "Site Code",
            "Site Name",
            "Associated Site Ids",
            "From",
            "To",
            "Gen Types",
            "CHP kWh",
            "LM kWh",
            "Turbine kWh",
            "PV kWh",
            "ccl-kwh",
            "ccl-rate",
            "ccl-gbp",
            "net-gbp",
            "vat-gbp",
            "gross-gbp",
            "sum-msp-kwh",
            "sum-msp-gbp",
            "problem",
        ],
        [
            "CI017",
            "Water Works",
            "",
            "2000-01-01 00:00",
            "2000-01-31 23:30",
            "turb",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "0.0",
            "0",
            "0.0",
            "0",
            "0.0",
            "",
        ],
    ]
    expected_str = "\n".join([",".join(v) for v in expected]) + "\n"
    assert expected_str == actual



glossary_intro = (
    "This is a list of the terms used in Chellow reports. The terms are often "
    "concatenated with a '-', eg site-code. Please get in touch if you have any "
    "improvements."
)

glossary_terms = {
    "creation-date": "Timstamp for when the item (report, issue etc.) was created",
    "site": "An area that supplies are attached to",
    "name": "Usually used as a suffix showing what something's called",
    "associated-site-ids": "An era is physically attached to a single site, but it "
    "can also feed other sites. The column 'associated-site-ids' are the ids of the "
    "other sites that the supply feeds.",
    "month": "A timestamp indicating the month that the row applies to",
    "metering-type": "Uses the dtc-meter-type to determine if an era is 'unmetered', "
    "'nhh', 'amr' or 'hh'.",
    "source": "The source of electricity for a supply. Can be the grid, a generator "
    "(gen) a submeter (sub) or a 3rd-party (from another electricity customer).",
    "generator-type": "If a source is a 'gen' then this column holds the type of "
    "generation: 'chp', 'pv', 'turb' or 'lm'.",
    "generated": "Electricity that's imported from a generator into the site.",
    "parasitic": "Electricity that's exported to a generator from the site.",
    "import": "Relates to electricity flowing through a supply towards the site. "
    "Often combined with other terms, eg. import-grid-kwh means electricity flowing "
    "into the site from the grid.",
    "imp": "Same as 'import'",
    "export": "Relates to electricity flowing through a supply away from the site. "
    "See 'import'.",
    "exp": "Same as 'export'",
    "displaced": "Electricity that's comes from generation and is used on site. "
    "It's calculated as displaced-kwh = imp-gen-kwh - exp-grid-kwh - exp-gen-kwh",
    "used": "The amount of electricity used on a site. The formula is used-kwh = "
    "imp-grid-kwh + imp-gen-kwh - exp-grid-kwh - exp-gen-kwh",
    "net-gbp": "The cost net of VAT.",
    "vat-gbp": "The VAT amount.",
    "gross-gbp": "The total of net + VAT",
    "billed": "An amount that comes from the actual bill from the supplier "
    "(or MOP / DC)",
    "supplier": "The retailer of electricity",
    "dc": "The appointed Data Collector responsible for metering data.",
    "mpan-core": "A sequence of 13 digits that uniquely identifies a supply. The "
    "first two digits tells you the DNO, the next 10 digits are generated by the "
    "DNO, and the final digit is a check-digit.",
    "contract": "Represents a legal contract for services for electricity supply, "
    "data collection or meter operation) that the customer has signed. The "
    "contract in Chellow contains a script that generates the virtual bill, and the "
    "contract has rate scripts attached to it which hold the rates for the "
    "contract.",
    "era": "The supply is made up of eras laid end to end together. So an era has a "
    "start date and a finish date, with the finish date possibly being 'ongoing'. "
    "An era holds the characteristics of a supply that may change over time, such "
    "as sc, imp_supplier, exp_supplier, pc, dc,  etc.",
    "msn": "Meter serial number. This changes each time there's a new meter. If you "
    "know the msn, you can look up the supply in ECOES.",
    "pc": "Profile class. Classifies the supply, where 00 means a half-hourly metered "
    "supply.",
    "energisation-status": "A supply can be energised, de-energised or disconnected. "
    "If a supply is disconnected that's represented in Chellow by giving the final "
    "era an end date, rather than 'ongoing'. If a supply is ongoing, then the "
    "energisation status can be 'D' for de-energised or 'E' for energised. A "
    "de-energised supply is physically still there and live, but out of use.",
    "gsp-group": "Grid supply point group. The physical area of the country that the "
    "supply is in.",
    "dno": "Distribution Network Operator. The organistion that operates the "
    "electricity network between the transmission system and the meter.",
    "voltage-level": "The voltage level of a supply can be LV, HV, or EHV",
    "is-substation": "Whether there's a substation on the site.",
    "llfc": "Line loss factor class. Determines the DUoS tariff.",
    "mop": "Meter operator. Responsible for installing, maintaining and removing "
    "electricity meters.",
    "settlement-status": "Can be 'settlement' or 'non-settlement'. If it's "
    "'settlement' then the supply is part of the national electricity system, but "
    "'non-settlement' means that it's local to the site (such as a generator supply "
    "or a sub-meter).",
    "comm": "The comms type",
    "cop": "Code of practice for the meter. Describes the capabilities of the meter.",
    "msp": "Meter supply point. The electricity consumption measured at the meter",
    "standing": "Tariff element that's a rolled up per-day charge",
    "availability": "The supply capacity agreed with the DNO in a connection "
    "agreement.",
    "kva": "The amount of apparent power (kVA).",
    "kw": "The amount of usable power (kW).",
    "kwh": "The amount of usable energy (kWh).",
    "kvar": "The amount of reactive power (kVAr).",
    "laf": "Loss adjustment multiplier. The losses going through the distribution "
    "network.",
    "gsp-kwh": "The amount of energy that has flowed through the grid supply point. "
    "It's calculated for each half-hour using gsp-kwh = laf x msp-kwh",
    "nbp-kwh": "Notional balancing point kwh. Aims to be the amount "
    "of electricity that has to be generated to meeting The demand at the meter, "
    "taking into account transmission and distribution losses. It's given for "
    "each half-hour by nbp-kwh = tlm * gsp-kwh.",
    "kvarh": "The amount of reactive energy (kVArh)",
    "tlm": "Transmission loss multiplier. Used to calculate the losses between the "
    "notional balancing point and the grid supply point. Depends on the half-hour and "
    "the gsp-group.",
}

glossary_elements = {
    "ccl": "Climate Change Levy.",
    "cm": "Capacity Market",
    "bsuos": "Balancing System Use of System",
    "standing": "Rolled up per-day charge",
    "duos": "Distribution Use of System. Tariff charges from the DNO.",
    "excess-availability": "This is the availability used above the agreed supply "
    "capacity.",
    "duos-fixed": "The part of DUoS that's per-day.",
    "duos-red": "The highest cost period that's charged per-kWh.",
    "duos-amber": "The medium DUoS cost period that's charged per-kwh.",
    "duos-green": "The lowest DUoS cost period that's charged per-kwh.",
    "duos-reactive": "The DUoS charge per-kvarh.",
    "aahedc": "Assistance for Areas with High Electricity Distribution Costs",
    "ro": "Renewables Obligation",
    "cfd": "Contracts for Difference",
    "triad": "NESO charge based on kW at three triad times",
    "tnuos": "Transmission costs. A per-day tariff based on bands "
    "derived from LLFCs.",
    "rab": "Regulated Asset Base. The nuclear power subsidy.",
    "duos-black": "DUoS tariff element for unmetered supplies.",
    "duos-yellow": "DUoS tariff element for unmetered supplies.",
    "obligation": "The bigger part of some tariffs, as opposed to 'operational'. "
    "Eg. cm-obligation-gbp",
    "operational": "The smaller part of some tariffs, as opposed to 'obligation'. "
    "Eg. cm-operational-gbp",
    "elexon": "For funding Elexon, charged at nbp-kwh",
}



import atexit
import collections
import csv
import threading
import traceback
from datetime import datetime as Datetime, timedelta
from io import StringIO

import requests

from sqlalchemy import select

from werkzeug.exceptions import BadRequest

from chellow.models import (
    Contract,
    MarketRole,
    Party,
    Session,
)
from chellow.utils import ct_datetime_now, hh_format, utc_datetime_now

importer = None

# https://bscdocs.elexon.co.uk/guidance-notes/isd-data-store-technical-specification

"""
def db_upgrade_52_to_53(sess, root_path):
    sess.execute(text("ALTER TABLE party ADD code CHARACTER VARYING;"))
    sess.execute(text("UPDATE party SET code = (name || cast(id as text));"))
    sess.execute(text("ALTER TABLE party ALTER code SET NOT NULL;"))
    sess.execute(
        text(
            "ALTER TABLE party ADD CONSTRAINT code_valid_from_key "
            "UNIQUE (code, valid_from);"
        )
    )

    sess.execute(text("ALTER TABLE market_role ADD mdd_code VARCHAR(4);"))
    sess.execute(text("UPDATE market_role SET mdd_code = code"))
    sess.execute(text("ALTER TABLE market_role ALTER COLUMN code TYPE VARCHAR(4);"))
"""


def _fetch_csv(s, entity_id):
    url = "https://datastore.helix.elexon.co.uk/api/ISD"
    params = {"filter": entity_id}
    res = s.get(url, params=params)
    res.raise_for_status()

    csv_file = StringIO(res.text)

    reader = csv.reader(csv_file)
    eid, entity_name, version_str, changed = next(reader)
    next(reader)
    assert entity_id == eid
    version = int(version_str)
    return reader, version


def _process_MarketRole(sess, log, reader):
    # DIP Market Role
    # DIP Role Description
    # Market Participant Role Code
    # Effective From Date {DIPRole}
    # Effective To Date {DIPRole}

    for values in reader:
        code = values[0]
        description = values[1]
        mdd_code = values[2]
        from_date_str = values[3]
        from_date = Datetime.fromisoformat(from_date_str)
        to_date_str = values[4]
        if to_date_str != "":
            raise BadRequest("Mapping finished")

        role = MarketRole.find_by_code(sess, code, from_date)
        if role is None:
            role = sess.scalars(
                select(MarketRole).where(MarketRole.mdd_code == mdd_code)
            ).one_or_none()
            if role is None:
                role = MarketRole.insert(sess, code, description, from_date, None)
            else:
                role.code = code
                sess.flush()
        else:
            if role.description != description:
                role.description = description
                sess.flush()


def _process_Party(sess, log, reader):
    # Market Participant ID String (4) Y
    # Market Participant Role Code String (1) N
    # DIP Participant ID String (10) Y
    # DIP Role Code String (5) Y
    # Effective From Date {MP2DPM} Date Y
    # Effective To Date {MP2DPM} Date N
    # Changed Int N

    for values in reader:
        participant_code = values[0]
        market_role_code = values[1]
        party_code = values[2]
        from_date_str = values[4]
        from_date = Datetime.fromisoformat(from_date_str)
        to_date_str = values[5]
        if to_date_str != "":
            raise BadRequest("Mapping finished")
        party = Party.get_by_participant_code_role_code(
            sess, participant_code, market_role_code, from_date
        )
        if party.code != party_code:
            log(f"Updating {party.code} with {party_code}.")


ENTITY_FUNCS = {
    "ISDEntityM18": _process_MarketRole,
    "ISDEntityM16": _process_Party,
}


def run_import(sess, log, set_progress, cur_version, s):
    log("Starting to import the ISD from Elexon Helix")
    version = 0
    return version
    for entity_code in ("ISDEntityM18", "ISDEntityM16"):
        reader, version = _fetch_csv(s, entity_code)
        if cur_version >= version:
            return cur_version
        ENTITY_FUNCS[entity_code](sess, log, reader)
    rows = [
        [
            "CIDC",
            "R",
            "01/01/2000",
            "",
            "Virtual",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "88",
        ],
        [
            "CROW",
            "R",
            "01/04/1996",
            "",
            "Non-settlement",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "99",
        ],
    ]
    _process_Party(sess, rows)

    return version


GLOBAL_ALERT = "There's a problem with an <a href='/e/isd'>ISD import</a>."
ISD_STATE_KEY = "isd"
LAST_RUN_KEY = "last_run"
VERSION_KEY = "version"
DELAY_DAYS = 1


class Isd(threading.Thread):
    def __init__(self):
        super().__init__(name="Industry Standing Data")
        self.messages = collections.deque(maxlen=500)
        self.progress = ""
        self.stopped = threading.Event()
        self.going = threading.Event()
        self.global_alert = None

    def stop(self):
        self.stopped.set()
        self.going.set()
        self.join()

    def go(self):
        self.going.set()

    def log(self, message):
        self.messages.appendleft(
            f"{ct_datetime_now().strftime('%Y-%m-%d %H:%M:%S')} - {message}"
        )

    def set_progress(self, progress):
        self.progress = progress

    def run(self):
        while not self.stopped.is_set():
            with Session() as sess:
                try:
                    config = Contract.get_non_core_by_name(sess, "configuration")
                    state = config.make_state()
                    isd_state = state.get(ISD_STATE_KEY, {})
                except BaseException as e:
                    msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                    self.log(f"{msg}{traceback.format_exc()}")
                    self.global_alert = GLOBAL_ALERT
                    sess.rollback()

            last_run = isd_state.get(LAST_RUN_KEY)
            if last_run is None or utc_datetime_now() - last_run > timedelta(
                days=DELAY_DAYS
            ):
                self.going.set()

            if self.going.is_set():
                self.global_alert = None
                with Session() as sess:
                    try:
                        config = Contract.get_non_core_by_name(sess, "configuration")
                        state = config.make_state()
                        try:
                            isd_state = state[ISD_STATE_KEY]
                        except KeyError:
                            isd_state = state[ISD_STATE_KEY] = {}

                        isd_state[LAST_RUN_KEY] = utc_datetime_now()
                        config.update_state(state)
                        sess.commit()
                        version = isd_state.get("version", 0)
                        s = requests.Session()
                        s.headers.update({"User-Agent": "Chellow"})

                        new_version = run_import(
                            sess, self.log, self.set_progress, version, s
                        )
                        isd_state["version"] = new_version
                        config.update_state(state)
                        sess.commit()
                    except BaseException as e:
                        msg = f"{e.description} " if isinstance(e, BadRequest) else ""
                        self.log(f"{msg}{traceback.format_exc()}")
                        self.global_alert = GLOBAL_ALERT
                        sess.rollback()
                    finally:
                        self.going.clear()
                        self.log("Finished importing ISD.")

            else:
                self.log(
                    f"The importer was last run at {hh_format(last_run)}. There will "
                    f"be another import when {DELAY_DAYS} days have elapsed since the "
                    f"last run."
                )
                self.going.wait(60 * 60)


def get_importer():
    return importer


def startup():
    global importer
    importer = Isd()
    importer.start()


@atexit.register
def shutdown():
    if importer is not None:
        importer.stop()
